#!/usr/bin/python
"""Class and helper functions for fitting the DeepN4 model."""

import logging

import numpy as np
from scipy.ndimage import gaussian_filter

from dipy.data import get_fnames
from dipy.nn.utils import normalize, recover_img, set_logger_level, transform_img
from dipy.testing.decorators import doctest_skip_parser
from dipy.utils.optpkg import optional_package

torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
if have_torch:
    from torch.nn import (
        Conv3d,
        ConvTranspose3d,
        InstanceNorm3d,
        LeakyReLU,
        MaxPool3d,
        Module,
        Sequential,
    )
else:

    class Module:
        pass

    logging.warning(
        "This model requires Pytorch.\
                    Please install these packages using \
                    pip."
    )

logging.basicConfig()
logger = logging.getLogger("deepn4")


class UNet3D(Module):
    def __init__(self, n_in, n_out):
        super(UNet3D, self).__init__()
        # Encoder
        c = 32
        self.ec0 = self.encoder_block(n_in, c, kernel_size=3, stride=1, padding=1)
        self.ec1 = self.encoder_block(c, c * 2, kernel_size=3, stride=1, padding=1)
        self.pool0 = MaxPool3d(2)
        self.ec2 = self.encoder_block(c * 2, c * 2, kernel_size=3, stride=1, padding=1)
        self.ec3 = self.encoder_block(c * 2, c * 4, kernel_size=3, stride=1, padding=1)
        self.pool1 = MaxPool3d(2)
        self.ec4 = self.encoder_block(c * 4, c * 4, kernel_size=3, stride=1, padding=1)
        self.ec5 = self.encoder_block(c * 4, c * 8, kernel_size=3, stride=1, padding=1)
        self.pool2 = MaxPool3d(2)
        self.ec6 = self.encoder_block(c * 8, c * 8, kernel_size=3, stride=1, padding=1)
        self.ec7 = self.encoder_block(c * 8, c * 16, kernel_size=3, stride=1, padding=1)
        self.el = Conv3d(c * 16, c * 16, kernel_size=1, stride=1, padding=0)

        # Decoder
        self.dc9 = self.decoder_block(
            c * 16, c * 16, kernel_size=2, stride=2, padding=0
        )
        self.dc8 = self.decoder_block(
            c * 16 + c * 8, c * 8, kernel_size=3, stride=1, padding=1
        )
        self.dc7 = self.decoder_block(c * 8, c * 8, kernel_size=3, stride=1, padding=1)
        self.dc6 = self.decoder_block(c * 8, c * 8, kernel_size=2, stride=2, padding=0)
        self.dc5 = self.decoder_block(
            c * 8 + c * 4, c * 4, kernel_size=3, stride=1, padding=1
        )
        self.dc4 = self.decoder_block(c * 4, c * 4, kernel_size=3, stride=1, padding=1)
        self.dc3 = self.decoder_block(c * 4, c * 4, kernel_size=2, stride=2, padding=0)
        self.dc2 = self.decoder_block(
            c * 4 + c * 2, c * 2, kernel_size=3, stride=1, padding=1
        )
        self.dc1 = self.decoder_block(c * 2, c * 2, kernel_size=3, stride=1, padding=1)
        self.dc0 = self.decoder_block(c * 2, n_out, kernel_size=1, stride=1, padding=0)
        self.dl = ConvTranspose3d(n_out, n_out, kernel_size=1, stride=1, padding=0)

    def encoder_block(self, in_channels, out_channels, kernel_size, stride, padding):
        layer = Sequential(
            Conv3d(
                in_channels,
                out_channels,
                kernel_size,
                stride=stride,
                padding=padding,
                bias=False,
            ),
            InstanceNorm3d(out_channels),
            LeakyReLU(),
        )
        return layer

    def decoder_block(self, in_channels, out_channels, kernel_size, stride, padding):
        layer = Sequential(
            ConvTranspose3d(
                in_channels,
                out_channels,
                kernel_size,
                stride=stride,
                padding=padding,
                bias=False,
            ),
            InstanceNorm3d(out_channels),
            LeakyReLU(),
        )
        return layer

    def forward(self, x):
        # Encodes
        e0 = self.ec0(x)
        syn0 = self.ec1(e0)
        del e0

        e1 = self.pool0(syn0)
        e2 = self.ec2(e1)
        syn1 = self.ec3(e2)
        del e1, e2

        e3 = self.pool1(syn1)
        e4 = self.ec4(e3)
        syn2 = self.ec5(e4)
        del e3, e4

        e5 = self.pool2(syn2)
        e6 = self.ec6(e5)
        e7 = self.ec7(e6)

        # Last layer without relu
        el = self.el(e7)
        del e5, e6, e7

        # Decode
        d9 = torch.cat((self.dc9(el), syn2), 1)
        del el, syn2

        d8 = self.dc8(d9)
        d7 = self.dc7(d8)
        del d9, d8

        d6 = torch.cat((self.dc6(d7), syn1), 1)
        del d7, syn1

        d5 = self.dc5(d6)
        d4 = self.dc4(d5)
        del d6, d5

        d3 = torch.cat((self.dc3(d4), syn0), 1)
        del d4, syn0

        d2 = self.dc2(d3)
        d1 = self.dc1(d2)
        del d3, d2

        d0 = self.dc0(d1)
        del d1

        # Last layer without relu
        out = self.dl(d0)
        return out


class DeepN4:
    """This class is intended for the DeepN4 model.

    The DeepN4 model :footcite:p:`Kanakaraj2024` predicts the bias field for
    magnetic field inhomogeneity correction on T1-weighted images.

    References
    ----------
    .. footbibliography::
    """

    @doctest_skip_parser
    def __init__(self, *, verbose=False, use_cuda=False):
        """Model initialization

        To obtain the pre-trained model, use fetch_default_weights() like:
        >>> deepn4_model = DeepN4() # skip if not have_torch
        >>> deepn4_model.fetch_default_weights() # skip if not have_torch

        This model is designed to take as input file T1 signal and predict
        bias field. Effectively, this model is mimicking bias correction.

        Parameters
        ----------
        verbose : bool, optional
            Whether to show information about the processing.
        use_cuda : bool, optional
            Whether to use GPU for processing.
            If False or no CUDA is detected, CPU will be used.
        """
        if not have_torch:
            raise torch()

        log_level = "INFO" if verbose else "CRITICAL"
        set_logger_level(log_level, logger)

        # DeepN4 network load

        self.model = UNet3D(1, 1)
        if use_cuda and torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")
        self.model = self.model.to(self.device)

    def fetch_default_weights(self):
        """Load the model pre-training weights to use for the fitting."""
        fetch_model_weights_path = get_fnames(name="deepn4_default_torch_weights")
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        """Load the custom pre-training weights to use for the fitting.

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights
        """
        try:
            self.model.load_state_dict(
                torch.load(
                    weights_path,
                    weights_only=True,
                    map_location=self.device,
                )["model_state_dict"]
            )
            self.model.eval()
        except ValueError as e:
            raise ValueError(
                "Expected input for the provided model weights \
                             do not match the declared model"
            ) from e

    def __predict(self, x_test):
        """Internal prediction function
        Predict bias field from input T1 signal

        Parameters
        ----------
        x_test : np.ndarray
            Image should match the required shape of the model.

        Returns
        -------
        np.ndarray (batch, ...)
            Predicted bias field
        """
        return self.model(x_test)[:, 0].detach().numpy()

    def pad(self, img, sz):
        tmp = np.zeros((sz, sz, sz))

        diff = int((sz - img.shape[0]) / 2)
        lx = max(diff, 0)
        lX = min(img.shape[0] + diff, sz)

        diff = (img.shape[0] - sz) / 2
        rx = max(int(np.floor(diff)), 0)
        rX = min(img.shape[0] - int(np.ceil(diff)), img.shape[0])

        diff = int((sz - img.shape[1]) / 2)
        ly = max(diff, 0)
        lY = min(img.shape[1] + diff, sz)

        diff = (img.shape[1] - sz) / 2
        ry = max(int(np.floor(diff)), 0)
        rY = min(img.shape[1] - int(np.ceil(diff)), img.shape[1])

        diff = int((sz - img.shape[2]) / 2)
        lz = max(diff, 0)
        lZ = min(img.shape[2] + diff, sz)

        diff = (img.shape[2] - sz) / 2
        rz = max(int(np.floor(diff)), 0)
        rZ = min(img.shape[2] - int(np.ceil(diff)), img.shape[2])

        tmp[lx:lX, ly:lY, lz:lZ] = img[rx:rX, ry:rY, rz:rZ]

        return tmp, [lx, lX, ly, lY, lz, lZ, rx, rX, ry, rY, rz, rZ]

    def load_resample(self, subj):
        input_data, [lx, lX, ly, lY, lz, lZ, rx, rX, ry, rY, rz, rZ] = self.pad(
            subj, 128
        )
        in_max = np.percentile(input_data[np.nonzero(input_data)], 99.99)
        input_data = normalize(input_data, min_v=0, max_v=in_max, new_min=0, new_max=1)
        input_data = np.squeeze(input_data)
        input_vols = np.zeros((1, 1, 128, 128, 128))
        input_vols[0, 0, :, :, :] = input_data

        return (
            torch.from_numpy(input_vols).float(),
            lx,
            lX,
            ly,
            lY,
            lz,
            lZ,
            rx,
            rX,
            ry,
            rY,
            rz,
            rZ,
            in_max,
        )

    def predict(self, img, img_affine, *, voxsize=(1, 1, 1), threshold=0.5):
        """Wrapper function to facilitate prediction of larger dataset.
        The function will mask, normalize, split, predict and 're-assemble'
        the data as a volume.

        Parameters
        ----------
        img : np.ndarray
            T1 image to predict and apply bias field
        img_affine : np.ndarray (4, 4)
            Affine matrix for the T1 image
        voxsize : np.ndarray or list or tuple (3,), optional
            voxel size of the T1 image.
        threshold : float, optional
            Threshold for cleaning the final correction field

        Returns
        -------
        final_corrected : np.ndarray (x, y, z)
            Predicted bias corrected image.
            The volume has matching shape to the input data
        """
        # Preprocess input data (resample, normalize, and pad)
        resampled_T1, inv_affine, mid_shape, offset_array, scale, crop_vs, pad_vs = (
            transform_img(img, img_affine, voxsize=voxsize)
        )
        (in_features, lx, lX, ly, lY, lz, lZ, rx, rX, ry, rY, rz, rZ, in_max) = (
            self.load_resample(resampled_T1)
        )

        # Run the model to get the bias field
        logfield = self.__predict(in_features.to(self.device))
        field = np.exp(logfield)
        field = field.squeeze()

        # Postprocess predicted field (reshape - unpad, smooth the field,
        # upsample)
        final_field = np.zeros(
            [resampled_T1.shape[0], resampled_T1.shape[1], resampled_T1.shape[2]]
        )
        final_field[rx:rX, ry:rY, rz:rZ] = field[lx:lX, ly:lY, lz:lZ]
        final_fields = gaussian_filter(final_field, sigma=3)
        upsample_final_field = recover_img(
            final_fields,
            inv_affine,
            mid_shape,
            img.shape,
            offset_array,
            voxsize,
            scale,
            crop_vs,
            pad_vs,
        )

        # Correct the image
        below_threshold_mask = np.abs(upsample_final_field) < threshold
        with np.errstate(divide="ignore", invalid="ignore"):
            final_corrected = np.where(
                below_threshold_mask, 0, img / upsample_final_field
            )

        return final_corrected



"""
=================================================================
Continuous and analytical diffusion signal modelling with MAP-MRI
=================================================================

We show how to model the diffusion signal as a linear combination of continuous
functions from the MAP-MRI basis :footcite:p:`Ozarslan2013`. This continuous
representation allows for the computation of many properties of both the signal
and diffusion propagator.

We show how to estimate the analytical orientation distribution function (ODF)
and a variety of scalar indices. These include rotationally invariant
quantities such as the mean squared displacement (MSD), Q-space inverse
variance (QIV), teturn-to-origin probability (RTOP) and non-Gaussianity (NG).
Interestingly, the MAP-MRI basis also allows for the computation of directional
indices, such as the return-to-axis probability (RTAP), the return-to-plane
probability (RTPP), and the parallel and perpendicular non-Gaussianity.

The estimation of these properties from noisy and sparse DWIs requires that the
fitting of the MAP-MRI basis is constrained and/or regularized. This can be
done by constraining the diffusion propagator to positive values
:footcite:p:`Ozarslan2013`, :footcite:p:`DelaHaije2020`, and through analytic
Laplacian regularization (MAPL) :footcite:p:`Fick2016b`.

First import the necessary modules:
"""

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst import mapmri
from dipy.viz import actor, window
from dipy.viz.plotting import compare_maps

###############################################################################
# Download and read the data for this tutorial.
#
# MAP-MRI requires multi-shell data, to properly fit the radial part of the
# basis. The total size of the downloaded data is 187.66 MBytes, however you
# only need to fetch it once.

fraw, fbval, fbvec, t1_fname = get_fnames(name="cfin_multib")

###############################################################################
# ``data`` contains the voxel data and ``gtab`` contains a ``GradientTable``
# object (gradient information e.g. b-values). For example, to show the
# b-values it is possible to write::
#
#    print(gtab.bvals)
#
# For the values of the q-space indices to make sense it is necessary to
# explicitly state the ``big_delta`` and ``small_delta`` parameters in the
# gradient table.

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)

big_delta = 0.0365  # seconds
small_delta = 0.0157  # seconds
gtab = gradient_table(
    bvals=gtab.bvals, bvecs=gtab.bvecs, big_delta=big_delta, small_delta=small_delta
)

data_small = data[40:65, 50:51]

print(f"data.shape {data.shape}")

###############################################################################
# The MAP-MRI Model can now be instantiated. The ``radial_order`` determines
# the expansion order of the basis, i.e., how many basis functions are used to
# approximate the signal.
#
# First, we must decide to use the anisotropic or isotropic MAP-MRI basis. As
# was shown in :footcite:p:`Fick2016b`, the isotropic basis is best used for
# tractography purposes, as the anisotropic basis has a bias towards smaller
# crossing angles in the ODF. For signal fitting and estimation of scalar
# quantities the anisotropic basis is suggested. The choice can be made by
# setting ``anisotropic_scaling=True`` or ``anisotropic_scaling=False``.
#
# First, we must select the method of regularization and/or constraining the
# basis fitting.
#
# - ``laplacian_regularization=True`` makes it use Laplacian regularization
#   :footcite:p:`Fick2016b`. This method essentially reduces spurious
#   oscillations in the reconstruction by minimizing the Laplacian of the fitted
#   signal. Several options can be given to select the regularization weight:
#
#     - ``regularization_weighting=GCV`` uses generalized cross-validation
#       :footcite:p:`Craven1979` to find an optimal weight.
#     - ``regularization_weighting=0.2`` for example would omit the GCV and
#       just set it to 0.2 (found to be reasonable in HCP data
#       :footcite:p:`Fick2016b`).
#     - ``regularization_weighting=np.array(weights)`` would make the GCV use
#       a custom range to find an optimal weight.
#
# - ``positivity_constraint=True`` makes it use a positivity constraint on the
#   diffusion propagator. This method constrains the final solution of the
#   diffusion propagator to be positive either globally
#   :footcite:p:`DelaHaije2020` or at a set of discrete points
#   :footcite:p:`Ozarslan2013`, since negative values should not exist.
#
#     - Setting ``global_constraints=True`` enforces positivity everywhere.
#       With the setting ``global_constraints=False`` positivity is enforced on
#       a grid determined by ``pos_grid`` and ``pos_radius``.
#
# Both methods do a good job of producing viable solutions to the signal
# fitting in practice. The difference is that the Laplacian regularization
# imposes smoothness over the entire signal, including the extrapolation
# beyond the measured signal. In practice this may result in, but does not
# guarantee, positive solutions of the diffusion propagator. The positivity
# constraint guarantees a positive solution which in general results in smooth
# solutions, but does not guarantee it.
#
# A suggested strategy is to use a low Laplacian weight together with a
# positivity constraint. In this way both desired properties are guaranteed
# in final solution. Higher Laplacian weights are recommended when the data is
# very noisy.
#
# We use the package CVXPY (https://www.cvxpy.org/) to solve convex
# optimization problems when ``positivity_constraint=True``, so we need to
# first install CVXPY. When using ``global_constraints=True`` to ensure
# global positivity, it is recommended to use the MOSEK solver
# (https://www.mosek.com/) together with CVXPY by setting
# ``cvxpy_solver='MOSEK'``. Different solvers can differ greatly in
# terms of runtime and solution accuracy, and in some cases solvers may show
# warnings about convergence or recommended option settings.
#
# For now we will generate the anisotropic models for different combinations.

radial_order = 6

map_model_laplacian_aniso = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    laplacian_regularization=True,
    laplacian_weighting=0.2,
)

map_model_positivity_aniso = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    laplacian_regularization=False,
    positivity_constraint=True,
)

map_model_both_aniso = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    laplacian_regularization=True,
    laplacian_weighting=0.05,
    positivity_constraint=True,
)

map_model_plus_aniso = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    laplacian_regularization=False,
    positivity_constraint=True,
    global_constraints=True,
)

###############################################################################
# Note that when we use only Laplacian regularization, the ``GCV`` option may
# select very low regularization weights in very anisotropic white matter such
# as the corpus callosum, resulting in corrupted scalar indices. In this
# example we therefore choose a preset value of 0.2, which was shown to be
# quite robust and also faster in practice :footcite:p:`Fick2016b`.
#
# We can then fit the MAP-MRI model to the data:

mapfit_laplacian_aniso = map_model_laplacian_aniso.fit(data_small)
mapfit_positivity_aniso = map_model_positivity_aniso.fit(data_small)
mapfit_both_aniso = map_model_both_aniso.fit(data_small)
mapfit_plus_aniso = map_model_plus_aniso.fit(data_small)

fits = [
    mapfit_laplacian_aniso,
    mapfit_positivity_aniso,
    mapfit_both_aniso,
    mapfit_plus_aniso,
]
fit_labels = ["MAPL", "CMAP", "CMAPL", "MAP+"]

###############################################################################
# From the fitted models we will first illustrate the estimation of q-space
# indices. We will compare the estimation using only Laplacian regularization
# (MAPL), using only the global positivity constraint (MAP+), using only
# positivity in discrete points (CMAP), or using both Laplacian regularization
# and positivity in discrete points (CMAPL). We first show the RTOP
# :footcite:p:`Ozarslan2013`.

compare_maps(
    fits,
    maps=["rtop"],
    fit_labels=fit_labels,
    map_labels=["RTOP"],
    filename="MAPMRI_rtop.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# It can be seen that all maps appear quite smooth and similar, although it is
# clear that the global positivity constraints provide smoother maps than the
# discrete constraints. Higher Laplacian weights also produce smoother maps,
# but tend to suppress the estimated RTOP values. The similarity and
# differences in reconstruction can be further illustrated by visualizing the
# analytic norm of the Laplacian of the fitted signal.

compare_maps(
    fits,
    maps=["norm_of_laplacian_signal"],
    fit_labels=fit_labels,
    map_labels=["Norm of Laplacian"],
    map_kwargs={"vmin": 0, "vmax": 3},
    filename="MAPMRI_norm_laplacian.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# A high Laplacian norm indicates that the gradient in the three-dimensional
# signal reconstruction changes a lot -- something that may indicate spurious
# oscillations. In the Laplacian reconstruction (left) we see that there are
# some isolated voxels that have a higher norm than the rest. In the positivity
# constrained reconstruction the norm is already smoother. When both methods
# are used together the overall norm gets smoother still, since both smoothness
# of the signal and positivity of the propagator are imposed.
#
# From now on we will just use the combined approach and the globally
# constrained approach, show all maps we can generate, and explain their
# significance.

fits = fits[2:]
fit_labels = fit_labels[2:]

compare_maps(
    fits,
    maps=["msd", "qiv", "rtop", "rtap", "rtpp"],
    fit_labels=fit_labels,
    map_labels=["MSD", "QIV", "RTOP", "RTAP", "RTPP"],
    filename="MAPMRI_maps.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# From left to right:
#
# - Mean squared displacement (MSD) is a measure for how far protons are able
#   to diffuse. a decrease in MSD indicates protons are hindered/restricted
#   more, as can be seen by the high MSD in the CSF, but low in the white
#   matter.
# - Q-space Inverse Variance (QIV) is a measure of variance in the signal,
#   which is said to have higher contrast to white matter than the MSD
#   :footcite:p:`Hosseinbor2013`. QIV has also been shown to have high
#   sensitivity to tissue composition change in a simulation study
#   :footcite:p:`Fick2016a`.
# - Return-to-origin probability (RTOP) quantifies the probability that a
#   proton will be at the same position at the first and second diffusion
#   gradient pulse. A higher RTOP indicates that the volume a spin is inside
#   of is smaller, meaning more overall restriction. This is illustrated by
#   the low values in CSF and high values in white matter.
# - Return-to-axis probability (RTAP) is a directional index that quantifies
#   the probability that a proton will be along the axis of the main
#   eigenvector of a diffusion tensor during both diffusion gradient pulses.
#   RTAP has been related to the apparent axon diameter
#   :footcite:p:`Ozarslan2013`, :footcite:p:`Fick2016b` under several strong
#   assumptions on the tissue composition and acquisition protocol.
# - Return-to-plane probability (RTPP) is a directional index that quantifies
#   the probability that a proton will be on the plane perpendicular to the
#   main eigenvector of a diffusion tensor during both gradient pulses. RTPP
#   is related to the length of a pore :footcite:p:`Ozarslan2013` but in
#   practice should be similar to that of Gaussian diffusion.
#
#
# It is also possible to estimate the amount of non-Gaussian diffusion in every
# voxel :footcite:p:`Ozarslan2013`. This quantity is estimated through the ratio
# between Gaussian volume (MAP-MRI's first basis function) and the non-Gaussian
# volume (all other basis functions) of the fitted signal. For this value to be
# physically meaningful we must use a b-value threshold in the MAP-MRI model.
# This threshold makes the scale estimation in MAP-MRI use only samples that
# realistically describe Gaussian diffusion, i.e., at low b-values.

map_model_both_ng = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    laplacian_regularization=True,
    laplacian_weighting=0.05,
    positivity_constraint=True,
    bval_threshold=2000,
)

map_model_plus_ng = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    positivity_constraint=True,
    global_constraints=True,
    bval_threshold=2000,
)

mapfit_both_ng = map_model_both_ng.fit(data_small)
mapfit_plus_ng = map_model_plus_ng.fit(data_small)

fits = [mapfit_both_ng, mapfit_plus_ng]
fit_labels = ["CMAPL", "MAP+"]

compare_maps(
    fits,
    maps=["ng", "ng_perpendicular", "ng_parallel"],
    fit_labels=fit_labels,
    map_labels=["NG", "NG perpendicular", "NG parallel"],
    filename="MAPMRI_ng.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# On the left we see the overall NG and on the right the directional
# perpendicular NG and parallel NG. The NG ranges from 1
# (completely non-Gaussian) to 0 (completely Gaussian). The overall NG of a
# voxel is always higher or equal than each of its components. It can be seen
# that NG has low values in the CSF and higher in the white matter.
#
# Increases or decreases in these values do not point to a specific
# microstructural change, but can indicate clues as to what is happening,
# similar to fractional anisotropy. An initial simulation study that quantifies
# the added value of q-space indices over DTI indices is given in
# :footcite:p:`Fick2016a`.
#
# The MAP-MRI framework also allows for the estimation of orientation
# distribution functions (ODFs). We recommend to use the isotropic
# implementation for this purpose, as the anisotropic implementation has a bias
# towards smaller crossing angles.
#
# For the isotropic basis we recommend to use a ``radial_order`` of 8, as the
# basis needs more generic and needs more basis functions to approximate the
# signal.

radial_order = 8
map_model_both_iso = mapmri.MapmriModel(
    gtab,
    radial_order=radial_order,
    laplacian_regularization=True,
    laplacian_weighting=0.1,
    positivity_constraint=True,
    anisotropic_scaling=False,
)

mapfit_both_iso = map_model_both_iso.fit(data_small)

###############################################################################
# Load an ODF reconstruction sphere.

sphere = get_sphere(name="repulsion724")

###############################################################################
# Compute the ODFs.
#
# The radial order ``s`` can be increased to sharpen the results, but it might
# also make the ODFs noisier. Always check the results visually.

odf = mapfit_both_iso.odf(sphere, s=2)
print(f"odf.shape {odf.shape}")

###############################################################################
# Display the ODFs.

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
sfu = actor.odf_slicer(odf, sphere=sphere, colormap="plasma", scale=0.5)
sfu.display(y=0)
sfu.RotateX(-90)
scene.add(sfu)
window.record(scene=scene, out_path="odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Orientation distribution functions (ODFs).
#
#
# References
# ----------
#
# .. footbibliography::
#



"""Various tools related to creating and working with streamlines.

This module provides tools for targeting streamlines using ROIs, for making
connectivity matrices from whole brain fiber tracking and some other tools that
allow streamlines to interact with image data.

Important Notes
-----------------
Dipy uses affine matrices to represent the relationship between streamline
points, which are defined as points in a continuous 3d space, and image voxels,
which are typically arranged in a discrete 3d grid. Dipy uses a convention
similar to nifti files to interpret these affine matrices. This convention is
that the point at the center of voxel ``[i, j, k]`` is represented by the point
``[x, y, z]`` where ``[x, y, z, 1] = affine * [i, j, k, 1]``.  Also when the
phrase "voxel coordinates" is used, it is understood to be the same as ``affine
= eye(4)``.

As an example, let's take a 2d image where the affine is::

    [[1., 0., 0.],
     [0., 2., 0.],
     [0., 0., 1.]]

The pixels of an image with this affine would look something like::

    A------------
    |   |   |   |
    | C |   |   |
    |   |   |   |
    ----B--------
    |   |   |   |
    |   |   |   |
    |   |   |   |
    -------------
    |   |   |   |
    |   |   |   |
    |   |   |   |
    ------------D

And the letters A-D represent the following points in
"real world coordinates"::

    A = [-.5, -1.]
    B = [ .5,  1.]
    C = [ 0.,  0.]
    D = [ 2.5,  5.]

"""

from collections import defaultdict
from functools import wraps
from itertools import combinations
from warnings import warn

from nibabel.affines import apply_affine
import numpy as np
from scipy.spatial.distance import cdist

from dipy.core.geometry import dist_to_corner
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking import metrics

# Import helper functions shared with vox2track
from dipy.tracking._utils import _mapping_to_voxel, _to_voxel_coordinates
from dipy.tracking.vox2track import _streamlines_in_mask


def density_map(streamlines, affine, vol_dims):
    """Count the number of unique streamlines that pass through each voxel.

    Parameters
    ----------
    streamlines : iterable
        A sequence of streamlines.
    affine : array_like (4, 4)
        The mapping from voxel coordinates to streamline points.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    vol_dims : 3 ints
        The shape of the volume to be returned containing the streamlines
        counts

    Returns
    -------
    image_volume : ndarray, shape=vol_dims
        The number of streamline points in each voxel of volume.

    Raises
    ------
    IndexError
        When the points of the streamlines lie outside of the return volume.

    Notes
    -----
    A streamline can pass through a voxel even if one of the points of the
    streamline does not lie in the voxel. For example a step from [0,0,0] to
    [0,0,2] passes through [0,0,1]. Consider subsegmenting the streamlines when
    the edges of the voxels are smaller than the steps of the streamlines.

    """
    lin_T, offset = _mapping_to_voxel(affine)
    counts = np.zeros(vol_dims, "int")
    for sl in streamlines:
        inds = _to_voxel_coordinates(sl, lin_T, offset)
        i, j, k = inds.T
        # this takes advantage of the fact that numpy's += operator only
        # acts once even if there are repeats in inds
        counts[i, j, k] += 1
    return counts


@warning_for_keywords()
def connectivity_matrix(
    streamlines,
    affine,
    label_volume,
    *,
    inclusive=False,
    symmetric=True,
    return_mapping=False,
    mapping_as_streamlines=False,
):
    """Count the streamlines that start and end at each label pair.

    Parameters
    ----------
    streamlines : sequence
        A sequence of streamlines.
    affine : array_like (4, 4)
        The mapping from voxel coordinates to streamline coordinates.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    label_volume : ndarray
        An image volume with an integer data type, where the intensities in the
        volume map to anatomical structures.
    inclusive: bool
        Whether to analyze the entire streamline, as opposed to just the
        endpoints. False by default.
    symmetric : bool, True by default
        Symmetric means we don't distinguish between start and end points. If
        symmetric is True, ``matrix[i, j] == matrix[j, i]``.
    return_mapping : bool, False by default
        If True, a mapping is returned which maps matrix indices to
        streamlines.
    mapping_as_streamlines : bool, False by default
        If True voxel indices map to lists of streamline objects. Otherwise
        voxel indices map to lists of integers.

    Returns
    -------
    matrix : ndarray
        The number of connection between each pair of regions in
        `label_volume`.
    mapping : defaultdict(list)
        ``mapping[i, j]`` returns all the streamlines that connect region `i`
        to region `j`. If `symmetric` is True mapping will only have one key
        for each start end pair such that if ``i < j`` mapping will have key
        ``(i, j)`` but not key ``(j, i)``.
    """

    # Error checking on label_volume
    kind = label_volume.dtype.kind
    labels_positive = (kind == "u") or ((kind == "i") and (label_volume.min() >= 0))
    valid_label_volume = labels_positive and label_volume.ndim == 3
    if not valid_label_volume:
        raise ValueError(
            "label_volume must be a 3d integer array with" "non-negative label values"
        )

    matrix = np.zeros(
        (np.max(label_volume) + 1, np.max(label_volume) + 1), dtype=np.int64
    )

    mapping = defaultdict(list)
    lin_T, offset = _mapping_to_voxel(affine)

    if inclusive:
        for i, sl in enumerate(streamlines):
            sl = _to_voxel_coordinates(sl, lin_T, offset)
            x, y, z = sl.T
            if symmetric:
                crossed_labels = np.unique(label_volume[x, y, z])
            else:
                crossed_labels = np.unique(label_volume[x, y, z], return_index=True)
                crossed_labels = crossed_labels[0][np.argsort(crossed_labels[1])]

            for comb in combinations(crossed_labels, 2):
                matrix[comb] += 1

                if return_mapping:
                    if mapping_as_streamlines:
                        mapping[comb].append(streamlines[i])
                    else:
                        mapping[comb].append(i)

    else:
        streamlines_end = np.array([sl[0 :: len(sl) - 1] for sl in streamlines])
        streamlines_end = _to_voxel_coordinates(streamlines_end, lin_T, offset)
        x, y, z = streamlines_end.T
        if symmetric:
            end_labels = np.sort(label_volume[x, y, z], axis=0)
        else:
            end_labels = label_volume[x, y, z]
        np.add.at(matrix, (end_labels[0].T, end_labels[1].T), 1)

        if return_mapping:
            if mapping_as_streamlines:
                for i, (a, b) in enumerate(end_labels.T):
                    mapping[a, b].append(streamlines[i])
            else:
                for i, (a, b) in enumerate(end_labels.T):
                    mapping[a, b].append(i)

    if symmetric:
        matrix = np.maximum(matrix, matrix.T)

    if return_mapping:
        return (matrix, mapping)
    else:
        return matrix


@warning_for_keywords()
def ndbincount(x, *, weights=None, shape=None):
    """Like bincount, but for nd-indices.

    Parameters
    ----------
    x : array_like (N, M)
        M indices to a an Nd-array
    weights : array_like (M,), optional
        Weights associated with indices
    shape : optional
        the shape of the output
    """
    x = np.asarray(x)
    if shape is None:
        shape = x.max(1) + 1

    x = np.ravel_multi_index(x, shape)
    out = np.bincount(x, weights, minlength=np.prod(shape))
    out.shape = shape

    return out


def reduce_labels(label_volume):
    """Reduce an array of labels to the integers from 0 to n with smallest
    possible n.

    Examples
    --------
    >>> labels = np.array([[1, 3, 9],
    ...                    [1, 3, 8],
    ...                    [1, 3, 7]])
    >>> new_labels, lookup = reduce_labels(labels)
    >>> lookup
    array([1, 3, 7, 8, 9])
    >>> new_labels #doctest: +ELLIPSIS
    array([[0, 1, 4],
           [0, 1, 3],
           [0, 1, 2]]...)
    >>> (lookup[new_labels] == labels).all()
    True

    """
    lookup_table = np.unique(label_volume)
    label_volume = lookup_table.searchsorted(label_volume)
    return label_volume, lookup_table


def subsegment(streamlines, max_segment_length):
    """Split the segments of the streamlines into small segments.

    Replaces each segment of each of the streamlines with the smallest possible
    number of equally sized smaller segments such that no segment is longer
    than max_segment_length. Among other things, this can useful for getting
    streamline counts on a grid that is smaller than the length of the
    streamline segments.

    Parameters
    ----------
    streamlines : sequence of ndarrays
        The streamlines to be subsegmented.
    max_segment_length : float
        The longest allowable segment length.

    Returns
    -------
    output_streamlines : generator
        A set of streamlines.

    Notes
    -----
    Segments of 0 length are removed. If unchanged

    Examples
    --------
    >>> streamlines = [np.array([[0,0,0],[2,0,0],[5,0,0]])]
    >>> list(subsegment(streamlines, 3.))
    [array([[ 0.,  0.,  0.],
           [ 2.,  0.,  0.],
           [ 5.,  0.,  0.]])]
    >>> list(subsegment(streamlines, 1))
    [array([[ 0.,  0.,  0.],
           [ 1.,  0.,  0.],
           [ 2.,  0.,  0.],
           [ 3.,  0.,  0.],
           [ 4.,  0.,  0.],
           [ 5.,  0.,  0.]])]
    >>> list(subsegment(streamlines, 1.6))
    [array([[ 0. ,  0. ,  0. ],
           [ 1. ,  0. ,  0. ],
           [ 2. ,  0. ,  0. ],
           [ 3.5,  0. ,  0. ],
           [ 5. ,  0. ,  0. ]])]

    """
    for sl in streamlines:
        diff = sl[1:] - sl[:-1]
        dist = np.sqrt((diff * diff).sum(-1))
        num_segments = np.ceil(dist / max_segment_length).astype("int")

        output_sl = np.empty((num_segments.sum() + 1, 3), "float")
        output_sl[0] = sl[0]

        count = 1
        for ii in range(len(num_segments)):
            ns = num_segments[ii]
            if ns == 1:
                output_sl[count] = sl[ii + 1]
                count += 1
            elif ns > 1:
                small_d = diff[ii] / ns
                point = sl[ii]
                for _ in range(ns):
                    point = point + small_d
                    output_sl[count] = point
                    count += 1
            elif ns == 0:
                pass
                # repeated point
            else:
                # this should never happen because ns should be a positive
                # int
                assert ns >= 0
        yield output_sl


@warning_for_keywords()
def seeds_from_mask(mask, affine, *, density=(1, 1, 1)):
    """Create seeds for fiber tracking from a binary mask.

    Seeds points are placed evenly distributed in all voxels of ``mask`` which
    are ``True``.

    Parameters
    ----------
    mask : binary 3d array_like
        A binary array specifying where to place the seeds for fiber tracking.
    affine : array, (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
        A seed point at the center the voxel ``[i, j, k]``
        will be represented as ``[x, y, z]`` where
        ``[x, y, z, 1] == np.dot(affine, [i, j, k , 1])``.
    density : int or array_like (3,)
        Specifies the number of seeds to place along each dimension. A
        ``density`` of `2` is the same as ``[2, 2, 2]`` and will result in a
        total of 8 seeds per voxel.

    See Also
    --------
    random_seeds_from_mask

    Raises
    ------
    ValueError
        When ``mask`` is not a three-dimensional array

    Examples
    --------
    >>> mask = np.zeros((3,3,3), 'bool')
    >>> mask[0,0,0] = 1
    >>> seeds_from_mask(mask, np.eye(4), density=[1,1,1])
    array([[ 0.,  0.,  0.]])

    """
    mask = np.asarray(mask, dtype=bool)
    if mask.ndim < 1:
        mask = np.reshape(mask, (3,))
    if mask.ndim != 3:
        raise ValueError("mask cannot be more than 3d")

    density = np.asarray(density, int)
    if density.size == 1:
        d = density
        density = np.empty(3, dtype=int)
        density.fill(d)
    elif density.shape != (3,):
        raise ValueError("density should be in integer array of shape (3,)")

    # Grid of points between -.5 and .5, centered at 0, with given density
    grid = np.mgrid[0 : density[0], 0 : density[1], 0 : density[2]]
    grid = grid.T.reshape((-1, 3))
    grid = grid / density
    grid += 0.5 / density - 0.5

    where = np.argwhere(mask)

    # Add the grid of points to each voxel in mask
    seeds = where[:, np.newaxis, :] + grid[np.newaxis, :, :]
    seeds = seeds.reshape((-1, 3))

    # Apply the spatial transform
    if seeds.any():
        # Use affine to move seeds into real world coordinates
        seeds = np.dot(seeds, affine[:3, :3].T)
        seeds += affine[:3, 3]

    return seeds


@warning_for_keywords()
def random_seeds_from_mask(
    mask, affine, *, seeds_count=1, seed_count_per_voxel=True, random_seed=None
):
    """Create randomly placed seeds for fiber tracking from a binary mask.

    Seeds points are placed randomly distributed in voxels of ``mask``
    which are ``True``.
    If ``seed_count_per_voxel`` is ``True``, this function is
    similar to ``seeds_from_mask()``, with the difference that instead of
    evenly distributing the seeds, it randomly places the seeds within the
    voxels specified by the ``mask``.

    Parameters
    ----------
    mask : binary 3d array_like
        A binary array specifying where to place the seeds for fiber tracking.
    affine : array, (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
        A seed point at the center the voxel ``[i, j, k]``
        will be represented as ``[x, y, z]`` where
        ``[x, y, z, 1] == np.dot(affine, [i, j, k , 1])``.
    seeds_count : int
        The number of seeds to generate. If ``seed_count_per_voxel`` is True,
        specifies the number of seeds to place in each voxel. Otherwise,
        specifies the total number of seeds to place in the mask.
    seed_count_per_voxel: bool
        If True, seeds_count is per voxel, else seeds_count is the total number
        of seeds.
    random_seed : int
        The seed for the random seed generator (numpy.random.Generator).

    See Also
    --------
    seeds_from_mask

    Raises
    ------
    ValueError
        When ``mask`` is not a three-dimensional array

    Examples
    --------
    >>> mask = np.zeros((3,3,3), 'bool')
    >>> mask[0,0,0] = 1
    >>> random_seeds_from_mask(mask, np.eye(4), seeds_count=1,
    ... seed_count_per_voxel=True, random_seed=1)
    array([[-0.23838787, -0.20150886,  0.31422574]])
    >>> random_seeds_from_mask(mask, np.eye(4), seeds_count=6,
    ... seed_count_per_voxel=True, random_seed=1)
    array([[-0.23838787, -0.20150886,  0.31422574],
           [-0.41435083, -0.26318949,  0.30127447],
           [ 0.44305611,  0.01132755,  0.47624371],
           [ 0.30500292,  0.30794079,  0.01532556],
           [ 0.03816435, -0.15672913, -0.13093276],
           [ 0.12509547,  0.3972138 ,  0.27568569]])
    >>> mask[0,1,2] = 1
    >>> random_seeds_from_mask(mask, np.eye(4),
    ... seeds_count=2, seed_count_per_voxel=True, random_seed=1)
    array([[ 0.30500292,  1.30794079,  2.01532556],
           [-0.23838787, -0.20150886,  0.31422574],
           [ 0.3702492 ,  0.78681721,  2.10314815],
           [-0.41435083, -0.26318949,  0.30127447]])

    """
    mask = np.asarray(mask, dtype=bool)
    if mask.ndim < 1:
        mask = np.reshape(mask, (3,))
    if mask.ndim != 3:
        raise ValueError("mask cannot be more than 3d")

    # Randomize the voxels
    rng = np.random.default_rng(random_seed)
    shape = mask.shape
    mask = mask.flatten()
    indices = np.arange(len(mask))
    rng.shuffle(indices)

    where = [np.unravel_index(i, shape) for i in indices if mask[i] == 1]
    num_voxels = len(where)

    if not seed_count_per_voxel:
        # Generate enough seeds per voxel
        seeds_per_voxel = seeds_count // num_voxels + 1
    else:
        seeds_per_voxel = seeds_count

    seeds = []
    for i in range(1, seeds_per_voxel + 1):
        for s in where:
            # Set the random seed with the current seed, the current value of
            # seeds per voxel and the global random seed.
            if random_seed is not None:
                s_random_seed = hash((np.sum(s) + 1) * i + random_seed) % (2**32 - 1)
                rng = np.random.default_rng(s_random_seed)
            # Generate random triplet
            grid = rng.random(3)
            seed = s + grid - 0.5
            seeds.append(seed)
    seeds = np.asarray(seeds)

    if not seed_count_per_voxel:
        # Select the requested amount
        seeds = seeds[:seeds_count]

    # Apply the spatial transform
    if seeds.any():
        # Use affine to move seeds into real world coordinates
        seeds = np.dot(seeds, affine[:3, :3].T)
        seeds += affine[:3, 3]

    return seeds


def _with_initialize(generator):
    """Allow one to write a generator with initialization code.

    All code up to the first yield is run as soon as the generator function is
    called and the first yield value is ignored.

    """

    @wraps(generator)
    def helper(*args, **kwargs):
        gen = generator(*args, **kwargs)
        next(gen)
        return gen

    return helper


@_with_initialize
@warning_for_keywords()
def target(streamlines, affine, target_mask, *, include=True):
    """Filter streamlines based on whether or not they pass through an ROI.

    Parameters
    ----------
    streamlines : iterable
        A sequence of streamlines. Each streamline should be a (N, 3) array,
        where N is the length of the streamline.
    affine : array (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    target_mask : array-like
        A mask used as a target. Non-zero values are considered to be within
        the target region.
    include : bool, default True
        If True, streamlines passing through `target_mask` are kept. If False,
        the streamlines not passing through `target_mask` are kept.

    Returns
    -------
    streamlines : generator
        A sequence of streamlines that pass through `target_mask`.

    Raises
    ------
    ValueError
        When the points of the streamlines lie outside of the `target_mask`.

    See Also
    --------
    density_map

    """
    target_mask = np.array(target_mask, dtype=bool, copy=True)
    lin_T, offset = _mapping_to_voxel(affine)
    yield
    # End of initialization

    for sl in streamlines:
        try:
            ind = _to_voxel_coordinates(sl, lin_T, offset)
            i, j, k = ind.T
            state = target_mask[i, j, k]
        except IndexError as e:
            raise ValueError("streamlines points are outside of target_mask") from e
        if state.any() == include:
            yield sl


@_with_initialize
@warning_for_keywords()
def target_line_based(streamlines, affine, target_mask, *, include=True):
    """Filter streamlines based on whether or not they pass through a ROI,
    using a line-based algorithm. Mostly used as a replacement of `target`
    for compressed streamlines.

    This function never returns single-point streamlines, whatever the
    value of `include`.

    See :footcite:`Bresenham1965` and :footcite:p:`Houde2015` for further
    details about the method.

    Parameters
    ----------
    streamlines : iterable
        A sequence of streamlines. Each streamline should be a (N, 3) array,
        where N is the length of the streamline.
    affine : array (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    target_mask : array-like
        A mask used as a target. Non-zero values are considered to be within
        the target region.
    include : bool, default True
        If True, streamlines passing through `target_mask` are kept. If False,
        the streamlines not passing through `target_mask` are kept.

    Returns
    -------
    streamlines : generator
        A sequence of streamlines that pass through `target_mask`.

    References
    ----------
    .. footbibliography::

    See Also
    --------
    dipy.tracking.utils.density_map
    dipy.tracking.streamline.compress_streamlines

    """
    target_mask = np.array(target_mask, dtype=np.uint8, copy=True)
    lin_T, offset = _mapping_to_voxel(affine)
    streamline_index = _streamlines_in_mask(streamlines, target_mask, lin_T, offset)
    yield
    # End of initialization

    for idx in np.where(streamline_index == [0, 1][include])[0]:
        yield streamlines[idx]


@warning_for_keywords()
def streamline_near_roi(streamline, roi_coords, tol, *, mode="any"):
    """Is a streamline near an ROI.

    Implements the inner loops of the :func:`near_roi` function.

    Parameters
    ----------
    streamline : array, shape (N, 3)
        A single streamline
    roi_coords : array, shape (M, 3)
        ROI coordinates transformed to the streamline coordinate frame.
    tol : float
        Distance (in the units of the streamlines, usually mm). If any
        coordinate in the streamline is within this distance from the center
        of any voxel in the ROI, this function returns True.
    mode : string
        One of {"any", "all", "either_end", "both_end"}, where return True
        if:

        "any" : any point is within tol from ROI.

        "all" : all points are within tol from ROI.

        "either_end" : either of the end-points is within tol from ROI

        "both_end" : both end points are within tol from ROI.

    Returns
    -------
    out : boolean

    """
    if not np.any(streamline):
        return False
    if len(roi_coords) == 0:
        return False
    if mode in ("any", "all"):
        s = streamline
    elif mode in ("either_end", "both_end"):
        # 'end' modes, use a streamline with 2 nodes:
        s = np.vstack([streamline[0], streamline[-1]])
    else:
        e_s = "For determining relationship to an array, you can use "
        e_s += "one of the following modes: 'any', 'all', 'both_end',"
        e_s += f"'either_end', but you entered: {mode}."
        raise ValueError(e_s)

    dist = cdist(s, roi_coords, "euclidean")

    if mode in ("any", "either_end"):
        return np.min(dist) <= tol
    else:
        return np.all(np.min(dist, -1) <= tol)


@warning_for_keywords()
def near_roi(streamlines, affine, region_of_interest, *, tol=None, mode="any"):
    """Provide filtering criteria for a set of streamlines based on whether
    they fall within a tolerance distance from an ROI.

    Parameters
    ----------
    streamlines : list or generator
        A sequence of streamlines. Each streamline should be a (N, 3) array,
        where N is the length of the streamline.
    affine : array (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    region_of_interest : ndarray
        A mask used as a target. Non-zero values are considered to be within
        the target region.
    tol : float
        Distance (in the units of the streamlines, usually mm). If any
        coordinate in the streamline is within this distance from the center
        of any voxel in the ROI, the filtering criterion is set to True for
        this streamline, otherwise False. Defaults to the distance between
        the center of each voxel and the corner of the voxel.
    mode : string, optional
        One of {"any", "all", "either_end", "both_end"}, where return True
        if:

        "any" : any point is within tol from ROI. Default.

        "all" : all points are within tol from ROI.

        "either_end" : either of the end-points is within tol from ROI

        "both_end" : both end points are within tol from ROI.

    Returns
    -------
    1D array of boolean dtype, shape (len(streamlines), )

    This contains `True` for indices corresponding to each streamline
    that passes within a tolerance distance from the target ROI, `False`
    otherwise.

    """
    dtc = dist_to_corner(affine)
    if tol is None:
        tol = dtc
    elif tol < dtc:
        w_s = "Tolerance input provided would create gaps in your"
        w_s += f" inclusion ROI. Setting to: {dtc}"
        warn(w_s, stacklevel=2)
        tol = dtc

    roi_coords = np.array(np.where(region_of_interest)).T
    x_roi_coords = apply_affine(affine, roi_coords)

    # If it's already a list, we can save time by pre-allocating the output
    if isinstance(streamlines, list):
        out = np.zeros(len(streamlines), dtype=bool)
        for ii, sl in enumerate(streamlines):
            out[ii] = streamline_near_roi(sl, x_roi_coords, tol=tol, mode=mode)
        return out
    # If it's a generator, we'll need to generate the output into a list
    else:
        out = []
        for sl in streamlines:
            out.append(streamline_near_roi(sl, x_roi_coords, tol=tol, mode=mode))

        return np.array(out, dtype=bool)


def length(streamlines):
    """Calculate the lengths of many streamlines in a bundle.

    Parameters
    ----------
    streamlines : list
        Each item in the list is an array with 3D coordinates of a streamline.

    Returns
    -------
    Iterator object which then computes the length of each
    streamline in the bundle, upon iteration.

    """
    return map(metrics.length, streamlines)


@warning_for_keywords()
def unique_rows(in_array, *, dtype="f4"):
    """Find the unique rows in an array.

    Parameters
    ----------
    in_array: ndarray
        The array for which the unique rows should be found

    dtype: str, optional
        This determines the intermediate representation used for the
        values. Should at least preserve the values of the input array.

    Returns
    -------
    u_return: ndarray
       Array with the unique rows of the original array.

    """
    # Sort input array
    order = np.lexsort(in_array.T)

    # Apply sort and compare neighbors
    x = in_array[order]
    diff_x = np.ones(len(x), dtype=bool)
    diff_x[1:] = (x[1:] != x[:-1]).any(-1)

    # Reverse sort and return unique rows
    un_order = order.argsort()
    diff_in_array = diff_x[un_order]
    return in_array[diff_in_array]


@_with_initialize
@warning_for_keywords()
def transform_tracking_output(tracking_output, affine, *, save_seeds=False):
    """Apply a linear transformation, given by affine, to streamlines.

    Parameters
    ----------
    tracking_output : Streamlines generator
        Either streamlines (list, ArraySequence) or a tuple with streamlines
        and seeds together
    affine : array (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    save_seeds : bool, optional
        If set, seeds associated to streamlines will be also moved and returned

    Returns
    -------
    streamlines : generator
        A generator for the sequence of transformed streamlines.
        If save_seeds is True, also return a generator for the
        transformed seeds.
    """
    lin_T = affine[:3, :3].T.copy()
    offset = affine[:3, 3].copy()
    yield
    # End of initialization

    if save_seeds:
        for sl, seed in tracking_output:
            yield np.dot(sl, lin_T) + offset, np.dot(seed, lin_T) + offset
    else:
        for sl in tracking_output:
            yield np.dot(sl, lin_T) + offset


def reduce_rois(rois, include):
    """Reduce multiple ROIs to one inclusion and one exclusion ROI.

    Parameters
    ----------
    rois : list or ndarray
        A list of 3D arrays, each with shape (x, y, z) corresponding to the
        shape of the brain volume, or a 4D array with shape (n_rois, x, y,
        z). Non-zeros in each volume are considered to be within the region.

    include : array or list
        A list or 1D array of boolean marking inclusion or exclusion
        criteria.

    Returns
    -------
    include_roi : boolean 3D array
        An array marking the inclusion mask.

    exclude_roi : boolean 3D array
        An array marking the exclusion mask

    Notes
    -----
    The include_roi and exclude_roi can be used to perform the operation: "(A
    or B or ...) and not (X or Y or ...)", where A, B are inclusion regions
    and X, Y are exclusion regions.

    """
    # throw warning if non bool roi detected
    if not np.all([irois.dtype == bool for irois in rois]):
        warn(
            "Non-boolean input mask detected. Treating all nonzeros as True.",
            stacklevel=2,
        )

    include_roi = np.zeros(rois[0].shape, dtype=bool)
    exclude_roi = np.zeros(rois[0].shape, dtype=bool)

    for i in range(len(rois)):
        if include[i]:
            include_roi |= rois[i] != 0
        else:
            exclude_roi |= rois[i] != 0

    return include_roi, exclude_roi


def _min_at(a, index, value):
    index = np.asarray(index)
    sort_keys = [value] + list(index)
    order = np.lexsort(sort_keys)
    index = index[:, order]
    value = value[order]
    uniq = np.ones(index.shape[1], dtype=bool)
    uniq[1:] = (index[:, 1:] != index[:, :-1]).any(axis=0)

    index = index[:, uniq]
    value = value[uniq]

    a[tuple(index)] = np.minimum(a[tuple(index)], value)


try:
    minimum_at = np.minimum.at
except AttributeError:
    minimum_at = _min_at


@warning_for_keywords()
def path_length(streamlines, affine, aoi, *, fill_value=-1):
    """Compute the shortest path, along any streamline, between aoi and
    each voxel.

    Parameters
    ----------
    streamlines : seq of (N, 3) arrays
        A sequence of streamlines, path length is given in mm along the curve
        of the streamline.
    aoi : array, 3d
        A mask (binary array) of voxels from which to start computing distance.
    affine : array (4, 4)
        The mapping between voxel indices and the point space for seeds.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    fill_value : float
        The value of voxel in the path length map that are not connected to the
        aoi.

    Returns
    -------
    plm : array
        Same shape as aoi. The minimum distance between every point and aoi
        along the path of a streamline.

    """
    aoi = np.asarray(aoi, dtype=bool)

    # path length map
    plm = np.empty(aoi.shape, dtype=float)
    plm[:] = np.inf
    lin_T, offset = _mapping_to_voxel(affine)
    for sl in streamlines:
        seg_ind = _to_voxel_coordinates(sl, lin_T, offset)
        i, j, k = seg_ind.T
        # Get where streamlines passes through aoi
        breaks = aoi[i, j, k]
        # Where streamline passes aoi, dist is zero
        i, j, k = seg_ind[breaks].T
        plm[i, j, k] = 0

        # If a streamline crosses aoi >1, re-start counting distance for each
        for seg in _as_segments(sl, breaks):
            i, j, k = _to_voxel_coordinates(seg[1:], lin_T, offset).T
            # Get the distance, in mm, between streamline points
            segment_length = np.sqrt(((seg[1:] - seg[:-1]) ** 2).sum(1))
            dist = segment_length.cumsum()
            # Updates path length map with shorter distances
            minimum_at(plm, (i, j, k), dist)
    if fill_value != np.inf:
        plm = np.where(plm == np.inf, fill_value, plm)
    return plm


def _part_segments(streamline, break_points):
    segments = np.split(streamline, break_points.nonzero()[0])
    # Skip first segment, all points before first break
    # first segment is empty when break_points[0] == 0
    segments = segments[1:]
    for each in segments:
        if len(each) > 1:
            yield each


def _as_segments(streamline, break_points):
    for seg in _part_segments(streamline, break_points):
        yield seg
    for seg in _part_segments(streamline[::-1], break_points[::-1]):
        yield seg


def max_angle_from_curvature(min_radius_curvature, step_size):
    """Get the maximum deviation angle from the minimum radius curvature.

    See :footcite:p:`Tournier2012` for further details about the method.

    Parameters
    ----------
    min_radius_curvature: float
        Minimum radius of curvature in mm.
    step_size: float
        The tracking step size in mm.

    Returns
    -------
    max_angle: float
        The maximum deviation angle in radian,
        given the radius curvature and the step size.

    References
    ----------
    .. footbibliography::

    """
    max_angle = 2.0 * np.arcsin(step_size / (2.0 * min_radius_curvature))
    if np.isnan(max_angle) or max_angle > np.pi / 2 or max_angle <= 0:
        w_msg = "The max_angle found is outside the interval [0 ; pi/2]."
        w_msg += "max_angle will be set to the default value pi/2"
        warn(w_msg, stacklevel=2)
        max_angle = np.pi / 2.0
    return max_angle


def min_radius_curvature_from_angle(max_angle, step_size):
    """Get minimum radius of curvature from a deviation angle.

    See :footcite:p:`Tournier2012` for further details about the method.

    Parameters
    ----------
    max_angle: float
        The maximum deviation angle in radian.
        theta should be between [0 - pi/2] otherwise default will be pi/2.
    step_size: float
        The tracking step size in mm.

    Returns
    -------
    min_radius_curvature: float
        Minimum radius of curvature in mm,
        given the maximum deviation angle theta and the step size.

    References
    ----------
    .. footbibliography::

    """
    if np.isnan(max_angle) or max_angle > np.pi / 2 or max_angle <= 0:
        w_msg = "The max_angle found is outside the interval [0 ; pi/2]."
        w_msg += "max_angle will be set to the default value pi/2"
        warn(w_msg, stacklevel=2)
        max_angle = np.pi / 2.0
    min_radius_curvature = step_size / 2 / np.sin(max_angle / 2)
    return min_radius_curvature


def seeds_directions_pairs(positions, peaks, *, max_cross=-1):
    """
    Pair each seed to the corresponding peaks. If multiple peaks are available
    the seed is repeated for each.

    Parameters
    ----------
    positions : array, (N, 3)
        Coordinates of the N positions.
    peaks : array (N, M, 3)
        Peaks at each position
    max_cross : int, optional
        The maximum number of direction to track from each seed in crossing
        voxels. By default all voxel peaks are used.

    Returns
    -------
    seeds : array (K, 3)
    directions : array (K, 3)
    """

    if (
        not len(positions.shape) == 2
        or not len(peaks.shape) == 3
        or not positions.shape[0] == peaks.shape[0]
        or not positions.shape[1] == 3
        or not peaks.shape[2] == 3
        or not peaks.shape[1] > 0
    ):
        raise ValueError(
            "The array shapes of the positions and peaks should"
            " be (N,3) and (N,M,3), respectively."
        )

    seeds = []
    directions = []

    for i, s in enumerate(positions):
        voxel_dirs_norm = np.linalg.norm(peaks[i, :, :], axis=1)
        voxel_dirs = (
            peaks[i, voxel_dirs_norm > 0, :]
            / voxel_dirs_norm[voxel_dirs_norm > 0, np.newaxis]
        )
        for d in voxel_dirs[:max_cross, :]:
            seeds.append(s)
            directions.append(d)

    return np.array(seeds), np.array(directions)



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False
from libc.stdio cimport printf
cimport numpy as cnp


cdef void take(
    double* odf,
    cnp.npy_intp* indices,
    cnp.npy_intp n_indices,
    double* values_out) noexcept nogil:
    """
    Mimic np.take(odf, indices) in Cython using pointers.

    Parameters
    ----------
    odf : double*
        Pointer to the input array from which values are taken.
    indices : npy_intp*
        Pointer to the array of indices specifying which elements to take.
    n_indices : npy_intp
        Number of indices to process.
    values_out : double*
        Pointer to the output array where the selected values will be stored.
    """
    cdef int i
    for i in range(n_indices):
        values_out[i] = odf[indices[i]]


cdef int where_to_insert(cnp.float_t* arr, cnp.float_t number, int size) noexcept nogil:
    cdef:
        int idx
        cnp.float_t current
    for idx in range(size - 1, -1, -1):
        current = arr[idx]
        if number >= current:
            return idx + 1

    return 0


cdef void cumsum(cnp.float_t* arr_in, cnp.float_t* arr_out, int N) noexcept nogil:
    cdef:
        cnp.npy_intp i = 0
        cnp.float_t csum = 0
    for i in range(N):
        csum += arr_in[i]
        arr_out[i] = csum


cdef void copy_point(double * a, double * b) noexcept nogil:
    cdef:
        cnp.npy_intp i = 0
    for i in range(3):
        b[i] = a[i]


cdef void scalar_muliplication_point(double * a, double scalar) noexcept nogil:
    cdef:
        cnp.npy_intp i = 0
    for i in range(3):
        a[i] *= scalar


cdef double norm(double * v) noexcept nogil:
    """Compute the vector norm.

    Parameters
    ----------
    v : double[3]
        input vector.

    """
    return sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])


cdef double dot(double * v1, double * v2) noexcept nogil:
    """Compute vectors dot product.

    Parameters
    ----------
    v1 : double[3]
        input vector 1.
    v2 : double[3]
        input vector 2.

    Returns
    -------
    _ : double
        dot product of input vectors.
    """
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


cdef void normalize(double * v) noexcept nogil:
    """Normalize the vector.

    Parameters
    ----------
    v : double[3]
        input vector

    Notes
    -----
    Overwrites the first argument.

    """
    cdef double scale = 1.0 / norm(v)
    v[0] = v[0] * scale
    v[1] = v[1] * scale
    v[2] = v[2] * scale


cdef void cross(double * out, double * v1, double * v2) noexcept nogil:
    """Compute vectors cross product.

    Parameters
    ----------
    out : double[3]
        output vector.
    v1 : double[3]
        input vector 1.
    v2 : double[3]
        input vector 2.

    Notes
    -----
    Overwrites the first argument.

    """
    out[0] = v1[1] * v2[2] - v1[2] * v2[1]
    out[1] = v1[2] * v2[0] - v1[0] * v2[2]
    out[2] = v1[0] * v2[1] - v1[1] * v2[0]


cdef void random_vector(double * out, RNGState* rng=NULL) noexcept nogil:
    """Generate a unit random vector

    Parameters
    ----------
    out : double[3]
        output vector

    Notes
    -----
    Overwrites the input
    """
    if rng == NULL:
        out[0] = 2.0 * random() - 1.0
        out[1] = 2.0 * random() - 1.0
        out[2] = 2.0 * random() - 1.0
    else:
        out[0] = 2.0 * random_float(rng) - 1.0
        out[1] = 2.0 * random_float(rng) - 1.0
        out[2] = 2.0 * random_float(rng) - 1.0
    normalize(out)


cdef void random_perpendicular_vector(double * out, double * v, RNGState* rng=NULL) noexcept nogil:
    """Generate a random perpendicular vector

    Parameters
    ----------
    out : double[3]
        output vector

    v : double[3]
        input vector

    Notes
    -----
    Overwrites the first argument
    """
    cdef double[3] tmp

    random_vector(tmp, rng)
    cross(out, v, tmp)
    normalize(out)


cdef (double, double) random_point_within_circle(double r, RNGState* rng=NULL) noexcept nogil:
    """Generate a random point within a circle

    Parameters
    ----------
    r : double
        The radius of the circle

    Returns
    -------
    x : double
        x coordinate of the random point

    y : double
        y coordinate of the random point

    """
    cdef double x = 1.0
    cdef double y = 1.0

    if rng == NULL:
        while (x * x + y * y) > 1:
            x = 2.0 * random() - 1.0
            y = 2.0 * random() - 1.0
    else:
        while (x * x + y * y) > 1:
            x = 2.0 * random_float(rng) - 1.0
            y = 2.0 * random_float(rng) - 1.0
    return (r * x, r * y)


cpdef double random() noexcept nogil:
    """Sample a random number between 0 and 1.

    Returns
    -------
    _ : double
        random number.
    """
    return rand() / float(RAND_MAX)


cpdef void seed(cnp.npy_uint32 s) noexcept nogil:
    """Set the random seed of stdlib.

    Parameters
    ----------
    s : int
        random seed.
    """
    srand(s)


cdef void print_c_array_pointer(double* arr, int size) noexcept nogil:
    cdef int i
    for i in range(size):
        printf("%f, ", arr[i])
    printf("\n\n\n")


cdef void seed_rng(RNGState* rng_state, cnp.npy_uint64 seed) noexcept nogil:
    """
    Seed the RNG state (thread-safe).

    Parameters
    ----------
    rng_state : RNGState*
        The RNG state.
    seed : int
        The seed value.

    """
    rng_state.state = 0
    rng_state.inc = (seed << 1) | 1
    next_rng(rng_state)  # Advance the state
    rng_state.state += seed
    next_rng(rng_state)


cdef cnp.npy_uint32 next_rng(RNGState* rng_state) noexcept nogil:
    """
    Generate the next random number (thread-safe, PCG algorithm).

    Parameters
    ----------
    rng_state : RNGState*
        The RNG state.

    Returns
    -------
    _ : int
        The next random number.
    """
    cdef cnp.npy_uint64 oldstate = rng_state.state
    rng_state.state = oldstate * 6364136223846793005ULL + rng_state.inc
    cdef cnp.npy_uint32 xorshifted = ((oldstate >> 18) ^ oldstate) >> 27
    cdef cnp.npy_uint32 rot = oldstate >> 59
    return (xorshifted >> rot) | (xorshifted << ((-rot) & 31))


cdef double random_float(RNGState* rng_state) noexcept nogil:
    """
    Generate a random float in [0, 1) (thread-safe).

    Parameters
    ----------
    rng_state : RNGState*
        The RNG state.

    Returns
    -------
    _ : float
        The random float.
    """
    return next_rng(rng_state) / <double>0xffffffff



"""
=================================================
Reconstruct with Diffusion Spectrum Imaging (DSI)
=================================================

We show how to apply Diffusion Spectrum Imaging :footcite:p:`Wedeen2005` to
diffusion MRI datasets of Cartesian keyhole diffusion gradients.

First import the necessary modules:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.core.ndindex import ndindex
from dipy.data import get_fnames, get_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.dsi import DiffusionSpectrumModel
from dipy.reconst.odf import gfa

###############################################################################
# Download and get the data filenames for this tutorial.

fraw, fbval, fbvec = get_fnames(name="taiwan_ntu_dsi")

###############################################################################
# img contains a nibabel Nifti1Image object (data) and gtab contains a
# GradientTable object (gradient information e.g. b-values). For example to
# read the b-values it is possible to write print(gtab.bvals).
#
# Load the raw diffusion data and the affine.

data, affine, voxel_size = load_nifti(fraw, return_voxsize=True)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
bvecs[1:] = bvecs[1:] / np.sqrt(np.sum(bvecs[1:] * bvecs[1:], axis=1))[:, None]
gtab = gradient_table(bvals, bvecs=bvecs)
print(f"data.shape {data.shape}")

###############################################################################
# This dataset has anisotropic voxel sizes, therefore reslicing is necessary.
#
# Instantiate the Model and apply it to the data.

dsmodel = DiffusionSpectrumModel(gtab)

###############################################################################
# Let's just use one slice only from the data.

dataslice = data[:, :, data.shape[2] // 2]

dsfit = dsmodel.fit(dataslice)

###############################################################################
# Load an odf reconstruction sphere

sphere = get_sphere(name="repulsion724")

###############################################################################
# Calculate the ODFs with this specific sphere

ODF = dsfit.odf(sphere)

print(f"ODF.shape {ODF.shape}")

###############################################################################
# In a similar fashion it is possible to calculate the PDFs of all voxels
# in one call with the following way

PDF = dsfit.pdf()

print(f"PDF.shape {PDF.shape}")

###############################################################################
# We see that even for a single slice this PDF array is close to 345 MBytes
# so we really have to be careful with memory usage when use this function
# with a full dataset.
#
# The simple solution is to generate/analyze the ODFs/PDFs by iterating
# through each voxel and not store them in memory if that is not necessary.

for index in ndindex(dataslice.shape[:2]):
    pdf = dsmodel.fit(dataslice[index]).pdf()

###############################################################################
# If you really want to save the PDFs of a full dataset on the disc we
# recommend using memory maps (``numpy.memmap``) but still have in mind that
# even if you do that for example for a dataset of volume size
# ``(96, 96, 60)`` you will need about 2.5 GBytes which can take less space
# when reasonable spheres (with < 1000 vertices) are used.
#
# Let's now calculate a map of Generalized Fractional Anisotropy (GFA)
# :footcite:p:`Tuch2004` using the DSI ODFs.

GFA = gfa(ODF)

fig_hist, ax = plt.subplots(1)
ax.set_axis_off()
plt.imshow(GFA.T)
plt.savefig("dsi_gfa.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# See also :ref:`sphx_glr_examples_built_reconstruction_reconst_dsi_metrics.py`
# for calculating different types of DSI maps.
#
#
# References
# ----------
#
# .. footbibliography::
#



from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.align.reslice import reslice
from dipy.align.tests.test_streamlinear import fornix_streamlines
from dipy.data import default_sphere, get_sphere, read_stanford_labels
from dipy.direction import peaks_from_model
from dipy.reconst.dti import color_fa, fractional_anisotropy
from dipy.reconst.shm import CsaOdfModel, descoteaux07_legacy_msg, sh_to_sf_matrix
from dipy.testing.decorators import set_random_number_generator, use_xvfb
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import center_streamlines, transform_streamlines
from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from dipy.viz import actor, colormap, window

skip_it = use_xvfb == "skip"


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator()
def test_slicer(rng):
    scene = window.Scene()
    data = 255 * rng.random((50, 50, 50))
    affine = np.diag([1, 3, 2, 1])
    data2, affine2 = reslice(data, affine, zooms=(1, 3, 2), new_zooms=(1, 1, 1))

    slicer = actor.slicer(data2, affine=affine2, interpolation="linear")
    slicer.display(x=None, y=None, z=25)

    scene.add(slicer)
    scene.reset_camera()
    scene.reset_clipping_range()

    # window.show(scene, reset_camera=False)
    arr = window.snapshot(scene, offscreen=True)
    report = window.analyze_snapshot(arr, find_objects=True)
    npt.assert_equal(report.objects, 1)
    npt.assert_array_equal([1, 3, 2] * np.array(data.shape), np.array(slicer.shape))


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_contour_from_roi():
    hardi_img, gtab, labels_img = read_stanford_labels()
    data = np.asanyarray(hardi_img.dataobj)
    labels = np.asanyarray(labels_img.dataobj)
    affine = hardi_img.affine

    white_matter = (labels == 1) | (labels == 2)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        csa_model = CsaOdfModel(gtab, sh_order_max=6)

        csa_peaks = peaks_from_model(
            csa_model,
            data,
            default_sphere,
            relative_peak_threshold=0.8,
            min_separation_angle=45,
            mask=white_matter,
        )

    classifier = ThresholdStoppingCriterion(csa_peaks.gfa, 0.25)

    seed_mask = labels == 2
    seeds = utils.seeds_from_mask(seed_mask, density=[1, 1, 1], affine=affine)

    # Initialization of LocalTracking.
    # The computation happens in the next step.
    streamlines = LocalTracking(csa_peaks, classifier, seeds, affine, step_size=2)

    # Compute streamlines and store as a list.
    streamlines = list(streamlines)

    # Prepare the display objects.
    streamlines_actor = actor.line(
        streamlines, colors=colormap.line_colors(streamlines)
    )
    seedroi_actor = actor.contour_from_roi(
        seed_mask, affine=affine, color=[0, 1, 1], opacity=0.5
    )

    with TemporaryDirectory() as out_dir:
        # Create the 3d display.
        sc = window.Scene()
        sc2 = window.Scene()
        sc.add(streamlines_actor)
        arr3 = window.snapshot(
            sc, fname=pjoin(out_dir, "test_surface3.png"), offscreen=True
        )
        report3 = window.analyze_snapshot(arr3, find_objects=True)
        sc2.add(streamlines_actor)
        sc2.add(seedroi_actor)
        arr4 = window.snapshot(
            sc2, fname=pjoin(out_dir, "test_surface4.png"), offscreen=True
        )
        report4 = window.analyze_snapshot(arr4, find_objects=True)

        # assert that the seed ROI rendering is not far
        # away from the streamlines (affine error)
        npt.assert_equal(report3.objects, report4.objects)
        # window.show(sc)
        # window.show(sc2)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator()
def test_bundle_maps(rng):
    scene = window.Scene()
    bundle = fornix_streamlines()
    bundle, _ = center_streamlines(bundle)

    mat = np.array([[1, 0, 0, 100], [0, 1, 0, 100], [0, 0, 1, 100], [0, 0, 0, 1.0]])

    bundle = transform_streamlines(bundle, mat)

    # metric = np.random.rand(*(200, 200, 200))
    metric = 100 * np.ones((200, 200, 200))

    # add lower values
    metric[100, :, :] = 100 * 0.5

    # create a nice orange-red colormap
    lut = actor.colormap_lookup_table(
        scale_range=(0.0, 100.0),
        hue_range=(0.0, 0.1),
        saturation_range=(1, 1),
        value_range=(1.0, 1),
    )

    line = actor.line(bundle, colors=metric, linewidth=0.1, lookup_colormap=lut)
    scene.add(line)
    scene.add(actor.scalar_bar(lookup_table=lut))

    report = window.analyze_scene(scene)

    npt.assert_almost_equal(report.actors, 1)
    # window.show(scene)

    scene.clear()

    nb_points = np.sum([len(b) for b in bundle])
    values = 100 * rng.random((nb_points))
    # values[:nb_points/2] = 0

    line = actor.streamtube(bundle, colors=values, linewidth=0.1, lookup_colormap=lut)
    scene.add(line)
    # window.show(scene)

    report = window.analyze_scene(scene)
    npt.assert_equal(report.actors_classnames[0], "vtkLODActor")

    scene.clear()

    colors = rng.random((nb_points, 3))
    # values[:nb_points/2] = 0

    line = actor.line(bundle, colors=colors, linewidth=2)
    scene.add(line)
    # window.show(scene)

    report = window.analyze_scene(scene)
    npt.assert_equal(report.actors_classnames[0], "vtkLODActor")
    # window.show(scene)

    arr = window.snapshot(scene)
    report2 = window.analyze_snapshot(arr)
    npt.assert_equal(report2.objects, 1)

    # try other input options for colors
    scene.clear()
    actor.line(bundle, colors=(1.0, 0.5, 0))
    actor.line(bundle, colors=np.arange(len(bundle)))
    actor.line(bundle)
    colors = [rng.random(b.shape) for b in bundle]
    actor.line(bundle, colors=colors)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_odf_slicer(interactive=False):
    # Prepare our data
    sphere = get_sphere(name="repulsion100")
    shape = (11, 11, 11, sphere.vertices.shape[0])
    odfs = np.ones(shape)

    affine = np.array(
        [
            [2.0, 0.0, 0.0, 3.0],
            [0.0, 2.0, 0.0, 3.0],
            [0.0, 0.0, 2.0, 1.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    mask = np.ones(odfs.shape[:3], bool)
    mask[:4, :4, :4] = False

    # Test that affine and mask work
    odf_actor = actor.odf_slicer(
        odfs, sphere=sphere, affine=affine, mask=mask, scale=0.25, colormap="blues"
    )

    k = 2
    _I, _J, _ = odfs.shape[:3]
    odf_actor.display_extent(0, _I - 1, 0, _J - 1, k, k)

    scene = window.Scene()
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()

    if interactive:
        window.show(scene, reset_camera=False)

    arr = window.snapshot(scene)
    report = window.analyze_snapshot(arr, find_objects=True)
    npt.assert_equal(report.objects, 11 * 11 - 16)

    # Test that global colormap works
    odf_actor = actor.odf_slicer(
        odfs,
        sphere=sphere,
        mask=mask,
        scale=0.25,
        colormap="blues",
        norm=False,
        global_cm=True,
    )
    scene.clear()
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()
    if interactive:
        window.show(scene)

    # Test that the most basic odf_slicer instantiation works
    odf_actor = actor.odf_slicer(odfs)
    scene.clear()
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()
    if interactive:
        window.show(scene)

    # Test that odf_slicer.display works properly
    scene.clear()
    scene.add(odf_actor)
    scene.add(actor.axes(scale=(11, 11, 11)))
    for i in range(11):
        odf_actor.display(x=i, y=None, z=None)
        if interactive:
            window.show(scene)
    for j in range(11):
        odf_actor.display(x=None, y=j, z=None)
        if interactive:
            window.show(scene)

    # With mask equal to zero everything should be black
    mask = np.zeros(odfs.shape[:3])
    odf_actor = actor.odf_slicer(
        odfs,
        sphere=sphere,
        mask=mask,
        scale=0.25,
        colormap="blues",
        norm=False,
        global_cm=True,
    )
    scene.clear()
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()
    if interactive:
        window.show(scene)

    # global_cm=True with colormap=None should raise an error
    npt.assert_raises(
        OSError,
        actor.odf_slicer,
        odfs,
        sphere=sphere,
        mask=None,
        scale=0.25,
        colormap=None,
        norm=False,
        global_cm=True,
    )

    # Dimension mismatch between sphere vertices and number
    # of SF coefficients will raise an error.
    npt.assert_raises(
        ValueError,
        actor.odf_slicer,
        odfs,
        mask=None,
        sphere=get_sphere(name="repulsion200"),
        scale=0.25,
    )

    # colormap=None and global_cm=False results in directionally encoded colors
    odf_actor = actor.odf_slicer(
        odfs,
        sphere=sphere,
        mask=None,
        scale=0.25,
        colormap=None,
        norm=False,
        global_cm=False,
    )
    scene.clear()
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()
    if interactive:
        window.show(scene)

    # Test that SH coefficients input works
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        B = sh_to_sf_matrix(sphere, sh_order_max=4, return_inv=False)
    odfs = np.zeros((11, 11, 11, B.shape[0]))
    odfs[..., 0] = 1.0
    odf_actor = actor.odf_slicer(odfs, sphere=sphere, B_matrix=B)

    scene.clear()
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()
    if interactive:
        window.show(scene)

    # Dimension mismatch between sphere vertices and dimension of
    # B matrix will raise an error.
    npt.assert_raises(
        ValueError,
        actor.odf_slicer,
        odfs,
        mask=None,
        sphere=get_sphere(name="repulsion200"),
    )

    # Test that constant colormap color works. Also test that sphere
    # normals are oriented correctly. Will show purple spheres with
    # a white contour.
    odf_contour = actor.odf_slicer(
        odfs, sphere=sphere, B_matrix=B, colormap=(255, 255, 255)
    )
    odf_contour.GetProperty().SetAmbient(1.0)
    odf_contour.GetProperty().SetFrontfaceCulling(True)

    odf_actor = actor.odf_slicer(
        odfs, sphere=sphere, B_matrix=B, colormap=(255, 0, 255), scale=0.4
    )
    scene.clear()
    scene.add(odf_contour)
    scene.add(odf_actor)
    scene.reset_camera()
    scene.reset_clipping_range()
    if interactive:
        window.show(scene)

    # Test that we can change the sphere on an active actor
    new_sphere = get_sphere(name="symmetric362")
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        new_B = sh_to_sf_matrix(new_sphere, sh_order_max=4, return_inv=False)
    odf_actor.update_sphere(new_sphere.vertices, new_sphere.faces, new_B)
    if interactive:
        window.show(scene)

    del odf_actor
    del odfs


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_tensor_slicer(interactive=False):
    evals = np.array([1.4, 0.35, 0.35]) * 10 ** (-3)
    evecs = np.eye(3)

    mevals = np.zeros((3, 2, 4, 3))
    mevecs = np.zeros((3, 2, 4, 3, 3))

    mevals[..., :] = evals
    mevecs[..., :, :] = evecs

    sphere = get_sphere(name="symmetric724")

    affine = np.eye(4)
    scene = window.Scene()

    tensor_actor = actor.tensor_slicer(
        mevals, mevecs, affine=affine, sphere=sphere, scale=0.3, opacity=0.4
    )
    _, J, K = mevals.shape[:3]
    scene.add(tensor_actor)
    scene.reset_camera()
    scene.reset_clipping_range()

    tensor_actor.display_extent(0, 1, 0, J, 0, K)
    if interactive:
        window.show(scene, reset_camera=False)

    tensor_actor.GetProperty().SetOpacity(1.0)
    if interactive:
        window.show(scene, reset_camera=False)

    npt.assert_equal(scene.GetActors().GetNumberOfItems(), 1)

    # Test extent
    big_extent = scene.GetActors().GetLastActor().GetBounds()
    big_extent_x = abs(big_extent[1] - big_extent[0])
    tensor_actor.display(x=2)

    if interactive:
        window.show(scene, reset_camera=False)

    small_extent = scene.GetActors().GetLastActor().GetBounds()
    small_extent_x = abs(small_extent[1] - small_extent[0])
    npt.assert_equal(big_extent_x > small_extent_x, True)

    # Test empty mask
    empty_actor = actor.tensor_slicer(
        mevals,
        mevecs,
        affine=affine,
        mask=np.zeros(mevals.shape[:3]),
        sphere=sphere,
        scale=0.3,
    )
    npt.assert_equal(empty_actor.GetMapper(), None)

    # Test mask
    mask = np.ones(mevals.shape[:3])
    mask[:2, :3, :3] = 0
    cfa = color_fa(fractional_anisotropy(mevals), mevecs)
    tensor_actor = actor.tensor_slicer(
        mevals,
        mevecs,
        affine=affine,
        mask=mask,
        scalar_colors=cfa,
        sphere=sphere,
        scale=0.3,
    )
    scene.clear()
    scene.add(tensor_actor)
    scene.reset_camera()
    scene.reset_clipping_range()

    if interactive:
        window.show(scene, reset_camera=False)

    mask_extent = scene.GetActors().GetLastActor().GetBounds()
    mask_extent_x = abs(mask_extent[1] - mask_extent[0])
    npt.assert_equal(big_extent_x > mask_extent_x, True)

    # test display
    tensor_actor.display()
    current_extent = scene.GetActors().GetLastActor().GetBounds()
    current_extent_x = abs(current_extent[1] - current_extent[0])
    npt.assert_equal(big_extent_x > current_extent_x, True)
    if interactive:
        window.show(scene, reset_camera=False)

    tensor_actor.display(y=1)
    current_extent = scene.GetActors().GetLastActor().GetBounds()
    current_extent_y = abs(current_extent[3] - current_extent[2])
    big_extent_y = abs(big_extent[3] - big_extent[2])
    npt.assert_equal(big_extent_y > current_extent_y, True)
    if interactive:
        window.show(scene, reset_camera=False)

    tensor_actor.display(z=1)
    current_extent = scene.GetActors().GetLastActor().GetBounds()
    current_extent_z = abs(current_extent[5] - current_extent[4])
    big_extent_z = abs(big_extent[5] - big_extent[4])
    npt.assert_equal(big_extent_z > current_extent_z, True)
    if interactive:
        window.show(scene, reset_camera=False)

    # Test error handling of the method when
    # incompatible dimension of mevals and evecs are passed.
    mevals = np.zeros((3, 2, 3))
    mevecs = np.zeros((3, 2, 4, 3, 3))

    with npt.assert_raises(RuntimeError):
        tensor_actor = actor.tensor_slicer(
            mevals,
            mevecs,
            affine=affine,
            mask=mask,
            scalar_colors=cfa,
            sphere=sphere,
            scale=0.3,
        )



import importlib
from inspect import getmembers, isfunction
import logging
import os
import sys
import warnings

import numpy as np
import trx.trx_file_memmap as tmm

from dipy.core.gradients import (
    extract_b0,
    extract_dwi_shell,
    gradient_table,
    mask_non_weighted_bvals,
)
from dipy.core.sphere import Sphere
from dipy.data import get_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti
from dipy.io.peaks import (
    load_pam,
    niftis_to_pam,
    pam_to_niftis,
    tensor_to_pam,
)
from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.reconst.shm import convert_sh_descoteaux_tournier
from dipy.reconst.utils import convert_tensors
from dipy.tracking.streamlinespeed import length
from dipy.utils.optpkg import optional_package
from dipy.utils.tractogram import concatenate_tractogram
from dipy.workflows.utils import handle_vol_idx
from dipy.workflows.workflow import Workflow

ne, have_ne, _ = optional_package("numexpr")


class IoInfoFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "io_info"

    def run(
        self,
        input_files,
        b0_threshold=50,
        bvecs_tol=0.01,
        bshell_thr=100,
        reference=None,
    ):
        """Provides useful information about different files used in
        medical imaging. Any number of input files can be provided. The
        program identifies the type of file by its extension.

        Parameters
        ----------
        input_files : variable string
            Any number of Nifti1, bvals or bvecs files.
        b0_threshold : float, optional
            Threshold used to find b0 volumes.
        bvecs_tol : float, optional
            Threshold used to check that norm(bvec) = 1 +/- bvecs_tol
            b-vectors are unit vectors.
        bshell_thr : float, optional
            Threshold for distinguishing b-values in different shells.
        reference : string, optional
            Reference anatomy for tck/vtk/fib/dpy file.
            support (.nii or .nii.gz).

        """
        np.set_printoptions(3, suppress=True)

        io_it = self.get_io_iterator()

        for input_path in io_it:
            mult_ = len(input_path)
            logging.info(f"-----------{mult_ * '-'}")
            logging.info(f"Looking at {input_path}")
            logging.info(f"-----------{mult_ * '-'}")

            ipath_lower = input_path.lower()
            extension = os.path.splitext(ipath_lower)[1]

            if ipath_lower.endswith(".nii") or ipath_lower.endswith(".nii.gz"):
                data, affine, img, vox_sz, affcodes = load_nifti(
                    input_path, return_img=True, return_voxsize=True, return_coords=True
                )
                logging.info(f"Data size {data.shape}")
                logging.info(f"Data type {data.dtype}")

                if data.ndim == 3:
                    logging.info(
                        f"Data min {data.min()} max {data.max()} avg {data.mean()}"
                    )
                    logging.info(
                        f"2nd percentile {np.percentile(data, 2)} "
                        f"98th percentile {np.percentile(data, 98)}"
                    )
                if data.ndim == 4:
                    logging.info(
                        f"Data min {data[..., 0].min()} "
                        f"max {data[..., 0].max()} "
                        f"avg {data[..., 0].mean()} of vol 0"
                    )
                    msg = (
                        f"2nd percentile {np.percentile(data[..., 0], 2)} "
                        f"98th percentile {np.percentile(data[..., 0], 98)} "
                        f"of vol 0"
                    )
                    logging.info(msg)
                logging.info(f"Native coordinate system {''.join(affcodes)}")
                logging.info(f"Affine Native to RAS matrix \n{affine}")
                logging.info(f"Voxel size {np.array(vox_sz)}")
                if np.sum(np.abs(np.diff(vox_sz))) > 0.1:
                    msg = "Voxel size is not isotropic. Please reslice.\n"
                    logging.warning(msg, stacklevel=2)

            if os.path.basename(input_path).lower().find("bval") > -1:
                bvals = np.loadtxt(input_path)
                logging.info(f"b-values \n{bvals}")
                logging.info(f"Total number of b-values {len(bvals)}")
                shells = np.sum(np.diff(np.sort(bvals)) > bshell_thr)
                logging.info(f"Number of gradient shells {shells}")
                logging.info(
                    f"Number of b0s {np.sum(bvals <= b0_threshold)} "
                    f"(b0_thr {b0_threshold})\n"
                )

            if os.path.basename(input_path).lower().find("bvec") > -1:
                bvecs = np.loadtxt(input_path)
                logging.info(f"Bvectors shape on disk is {bvecs.shape}")
                rows, cols = bvecs.shape
                if rows < cols:
                    bvecs = bvecs.T
                logging.info(f"Bvectors are \n{bvecs}")
                norms = np.array([np.linalg.norm(bvec) for bvec in bvecs])
                res = np.where((norms <= 1 + bvecs_tol) & (norms >= 1 - bvecs_tol))
                ncl1 = np.sum(norms < 1 - bvecs_tol)
                logging.info(f"Total number of unit bvectors {len(res[0])}")
                logging.info(f"Total number of non-unit bvectors {ncl1}\n")

            if extension in [".trk", ".tck", ".trx", ".vtk", ".vtp", ".fib", ".dpy"]:
                sft = None
                if extension in [".trk", ".trx"]:
                    sft = load_tractogram(input_path, "same", bbox_valid_check=False)
                else:
                    sft = load_tractogram(input_path, reference, bbox_valid_check=False)

                lengths_mm = list(length(sft.streamlines))

                sft.to_voxmm()

                lengths, steps = [], []
                for streamline in sft.streamlines:
                    lengths += [len(streamline)]
                    steps += [np.sqrt(np.sum(np.diff(streamline, axis=0) ** 2, axis=1))]
                steps = np.hstack(steps)

                logging.info(f"Number of streamlines: {len(sft)}")
                logging.info(f"min_length_mm: {float(np.min(lengths_mm))}")
                logging.info(f"mean_length_mm: {float(np.mean(lengths_mm))}")
                logging.info(f"max_length_mm: {float(np.max(lengths_mm))}")
                logging.info(f"std_length_mm: {float(np.std(lengths_mm))}")
                logging.info(f"min_length_nb_points: {float(np.min(lengths))}")
                logging.info("mean_length_nb_points: " f"{float(np.mean(lengths))}")
                logging.info(f"max_length_nb_points: {float(np.max(lengths))}")
                logging.info(f"std_length_nb_points: {float(np.std(lengths))}")
                logging.info(f"min_step_size: {float(np.min(steps))}")
                logging.info(f"mean_step_size: {float(np.mean(steps))}")
                logging.info(f"max_step_size: {float(np.max(steps))}")
                logging.info(f"std_step_size: {float(np.std(steps))}")
                logging.info(
                    "data_per_point_keys: " f"{list(sft.data_per_point.keys())}"
                )
                logging.info(
                    "data_per_streamline_keys: "
                    f"{list(sft.data_per_streamline.keys())}"
                )

        np.set_printoptions()


class FetchFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "fetch"

    @staticmethod
    def get_fetcher_datanames():
        """Gets available dataset and function names.

        Returns
        -------
        available_data: dict
            Available dataset and function names.

        """

        fetcher_module = FetchFlow.load_module("dipy.data.fetcher")

        available_data = dict(
            {
                (name.replace("fetch_", ""), func)
                for name, func in getmembers(fetcher_module, isfunction)
                if name.lower().startswith("fetch_")
                and func is not fetcher_module.fetch_data
            }
        )

        return available_data

    @staticmethod
    def load_module(module_path):
        """Load / reload an external module.

        Parameters
        ----------
        module_path: string
            the path to the module relative to the main script

        Returns
        -------
        module: module object

        """
        if module_path in sys.modules:
            return importlib.reload(sys.modules[module_path])
        else:
            return importlib.import_module(module_path)

    def run(
        self,
        data_names,
        subjects=None,
        include_optional=False,
        include_afq=False,
        hcp_bucket="hcp-openaccess",
        hcp_profile_name="hcp",
        hcp_study="HCP_1200",
        hcp_aws_access_key_id=None,
        hcp_aws_secret_access_key=None,
        out_dir="",
    ):
        """Download files to folder and check their md5 checksums.

        To see all available datasets, please type "list" in data_names.

        Parameters
        ----------
        data_names : variable string
            Any number of Nifti1, bvals or bvecs files.
        subjects : variable string, optional
            Identifiers of the subjects to download. Used only by the HBN & HCP dataset.
            For example with HBN dataset: --subject NDARAA948VFH NDAREK918EC2
        include_optional : bool, optional
            Include optional datasets.
        include_afq : bool, optional
            Whether to include pyAFQ derivatives. Used only by the HBN dataset.
        hcp_bucket : string, optional
            The name of the HCP S3 bucket.
        hcp_profile_name : string, optional
            The name of the AWS profile used for access.
        hcp_study : string, optional
            Which HCP study to grab.
        hcp_aws_access_key_id : string, optional
            AWS credentials to HCP AWS S3. Will only be used if `profile_name` is
            set to False.
        hcp_aws_secret_access_key : string, optional
            AWS credentials to HCP AWS S3. Will only be used if `profile_name` is
            set to False.
        out_dir : string, optional
            Output directory.

        """
        if out_dir:
            dipy_home = os.environ.get("DIPY_HOME", None)
            os.environ["DIPY_HOME"] = out_dir

        available_data = FetchFlow.get_fetcher_datanames()

        data_names = [name.lower() for name in data_names]

        if "all" in data_names:
            logging.warning("Skipping HCP and HBN datasets.")
            available_data.pop("hcp", None)
            available_data.pop("hbn", None)
            for name, fetcher_function in available_data.items():
                if name in ["hcp", "hbn"]:
                    continue
                logging.info("------------------------------------------")
                logging.info(f"Fetching at {name}")
                logging.info("------------------------------------------")
                fetcher_function(include_optional=include_optional)

        elif "list" in data_names:
            logging.info(
                "Please, select between the following data names: \n"
                f"{', '.join(available_data.keys())}"
            )

        else:
            skipped_names = []
            for data_name in data_names:
                if data_name not in available_data.keys():
                    skipped_names.append(data_name)
                    continue

                logging.info("------------------------------------------")
                logging.info(f"Fetching at {data_name}")
                logging.info("------------------------------------------")
                if data_name == "hcp":
                    if not subjects:
                        logging.error(
                            "Please provide the subjects to download the HCP dataset."
                        )
                        continue
                    try:
                        available_data[data_name](
                            subjects=subjects,
                            bucket=hcp_bucket,
                            profile_name=hcp_profile_name,
                            study=hcp_study,
                            aws_access_key_id=hcp_aws_access_key_id,
                            aws_secret_access_key=hcp_aws_secret_access_key,
                        )
                    except Exception as e:
                        logging.error(
                            f"Error while fetching HCP dataset: {str(e)}", exc_info=True
                        )
                elif data_name == "hbn":
                    if not subjects:
                        logging.error(
                            "Please provide the subjects to download the HBN dataset."
                        )
                        continue
                    try:
                        available_data[data_name](
                            subjects=subjects, include_afq=include_afq
                        )
                    except Exception as e:
                        logging.error(
                            f"Error while fetching HBN dataset: {str(e)}", exc_info=True
                        )
                else:
                    available_data[data_name](include_optional=include_optional)

            nb_success = len(data_names) - len(skipped_names)
            print("\n")
            logging.info(f"Fetched {nb_success} / {len(data_names)} Files ")
            if skipped_names:
                logging.warn(f"Skipped data name(s): {' '.join(skipped_names)}")
                logging.warn(
                    "Please, select between the following data names: "
                    f"{', '.join(available_data.keys())}"
                )

        if out_dir:
            if dipy_home:
                os.environ["DIPY_HOME"] = dipy_home
            else:
                os.environ.pop("DIPY_HOME", None)

            # We load the module again so that if we run another one of these
            # in the same process, we don't have the env variable pointing
            # to the wrong place
            self.load_module("dipy.data.fetcher")


class SplitFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "split"

    def run(self, input_files, vol_idx=0, out_dir="", out_split="split.nii.gz"):
        """Splits the input 4D file and extracts the required 3D volume.

        Parameters
        ----------
        input_files : variable string
            Any number of Nifti1 files
        vol_idx : int, optional
            Index of the 3D volume to extract.
        out_dir : string, optional
            Output directory.
        out_split : string, optional
            Name of the resulting split volume

        """
        io_it = self.get_io_iterator()
        for fpath, osplit in io_it:
            logging.info(f"Splitting {fpath}")
            data, affine, image = load_nifti(fpath, return_img=True)

            if vol_idx == 0:
                logging.info("Splitting and extracting 1st b0")

            split_vol = data[..., vol_idx]
            save_nifti(osplit, split_vol, affine, hdr=image.header)

            logging.info(f"Split volume saved as {osplit}")


class ExtractB0Flow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "extract_b0"

    def run(
        self,
        input_files,
        bvalues_files,
        b0_threshold=50,
        group_contiguous_b0=False,
        strategy="mean",
        out_dir="",
        out_b0="b0.nii.gz",
    ):
        """Extract on or multiple b0 volume from the input 4D file.

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bvalues_files : string
            Path to the bvalues files. This path may contain wildcards to use
            multiple bvalues files at once.
        b0_threshold : float, optional
            Threshold used to find b0 volumes.
        group_contiguous_b0 : bool, optional
            If True, each contiguous b0 volumes are grouped together.
        strategy : str, optional
            The extraction strategy, of either:

                - first: select the first b0 found.
                - all: select them all.
                - mean: average them.

            When used in conjunction with the batch parameter set to True, the
            strategy is applied individually on each continuous set found.
        out_dir : string, optional
            Output directory.
        out_b0 : string, optional
            Name of the resulting b0 volume.

        """
        io_it = self.get_io_iterator()
        for dwi, bval, ob0 in io_it:
            logging.info("Extracting b0 from {0}".format(dwi))
            data, affine, image = load_nifti(dwi, return_img=True)

            bvals, bvecs = read_bvals_bvecs(bval, None)
            # If all b-values are smaller or equal to the b0 threshold, it is
            # assumed that no thresholding is requested
            if any(mask_non_weighted_bvals(bvals, b0_threshold)):
                if b0_threshold < bvals.min():
                    warnings.warn(
                        f"b0_threshold (value: {b0_threshold}) is too low, "
                        "increase your b0_threshold. It should be higher than the "
                        f"first b0 value ({bvals.min()}).",
                        stacklevel=2,
                    )

            bvecs = np.random.randn(bvals.shape[0], 3)
            norms = np.linalg.norm(bvecs, axis=1, keepdims=True)
            bvecs = bvecs / norms
            gtab = gradient_table(bvals, bvecs=bvecs, b0_threshold=b0_threshold)
            b0s_result = extract_b0(
                data,
                gtab.b0s_mask,
                group_contiguous_b0=group_contiguous_b0,
                strategy=strategy,
            )

            if b0s_result.ndim == 3:
                save_nifti(ob0, b0s_result, affine, hdr=image.header)
                logging.info("b0 saved as {0}".format(ob0))
            elif b0s_result.ndim == 4:
                for i in range(b0s_result.shape[-1]):
                    save_nifti(
                        ob0.replace(".nii", f"_{i}.nii"),
                        b0s_result[..., i],
                        affine,
                        hdr=image.header,
                    )
                    logging.info(
                        "b0 saved as {0}".format(ob0.replace(".nii", f"_{i}.nii"))
                    )
            else:
                logging.error("No b0 volumes found")


class ExtractShellFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "extract_shell"

    def run(
        self,
        input_files,
        bvalues_files,
        bvectors_files,
        bvals_to_extract=None,
        b0_threshold=50,
        bvecs_tol=0.01,
        tol=20,
        group_shells=True,
        out_dir="",
        out_shell="shell.nii.gz",
    ):
        """Extract shells from the input 4D file.

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bvalues_files : string
            Path to the bvalues files. This path may contain wildcards to use
            multiple bvalues files at once.
        bvectors_files : string
            Path to the bvectors files. This path may contain wildcards to use
            multiple bvectors files at once.
        bvals_to_extract : string, optional
            List of b-values to extract. You can provide a single b-values or a range
            of b-values separated by a dash. For example, to extract b-values 0, 1,
            and 2, you can use '0-2'. You can also provide a list of b-values separated
            by a comma. For example, to extract b-values 0, 1, 2, 8, 10, 11 and 12,
            you can use '0-2,8,10-12'.
        b0_threshold : float, optional
            Threshold used to find b0 volumes.
        bvecs_tol : float, optional
            Threshold used to check that norm(bvec) = 1 +/- bvecs_tol
        tol : int, optional
            Tolerance range for b-value selection. A value of 20 means volumes with
            b-values within ±20 units of the specified b-values will be extracted.
        group_shells : bool, optional
            If True, extracted volumes are grouped into a single array. If False,
            returns a list of separate volumes.
        out_dir : string, optional
            Output directory.
        out_shell : string, optional
            Name of the resulting shell volume.

        """
        io_it = self.get_io_iterator()
        if bvals_to_extract is None:
            logging.error(
                "Please provide a list of b-values to extract."
                " e.g: --bvals_to_extract 1000 2000 3000"
            )
            sys.exit(1)

        bvals_to_extract = handle_vol_idx(bvals_to_extract)

        for dwi, bval, bvec, oshell in io_it:
            logging.info("Extracting shell from {0}".format(dwi))
            data, affine, image = load_nifti(dwi, return_img=True)

            bvals, bvecs = read_bvals_bvecs(bval, bvec)
            # If all b-values are smaller or equal to the b0 threshold, it is
            # assumed that no thresholding is requested
            if any(mask_non_weighted_bvals(bvals, b0_threshold)):
                if b0_threshold < bvals.min():
                    warnings.warn(
                        f"b0_threshold (value: {b0_threshold}) is too low, "
                        "increase your b0_threshold. It should be higher than the "
                        f"first b0 value ({bvals.min()}).",
                        stacklevel=2,
                    )
            gtab = gradient_table(
                bvals, bvecs=bvecs, b0_threshold=b0_threshold, atol=bvecs_tol
            )
            indices, shell_data, output_bvals, output_bvecs = extract_dwi_shell(
                data,
                gtab,
                bvals_to_extract,
                tol=tol,
                group_shells=group_shells,
            )

            for i, shell in enumerate(shell_data):
                shell_value = np.unique(output_bvals[i]).astype(int).astype(str)
                shell_value = "_".join(shell_value.tolist())
                save_nifti(
                    oshell.replace(".nii", f"_{shell_value}.nii"),
                    shell,
                    affine,
                    hdr=image.header,
                )
                logging.info(
                    "b0 saved as {0}".format(
                        oshell.replace(".nii", f"_{shell_value}.nii")
                    )
                )


class ExtractVolumeFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "extract_volume"

    def run(
        self, input_files, vol_idx=0, grouped=True, out_dir="", out_vol="volume.nii.gz"
    ):
        """Extracts the required volume from the input 4D file.

        Parameters
        ----------
        input_files : string
            Any number of Nifti1 files
        vol_idx : string, optional
            Indexes of the 3D volume to extract. Index start from 0. You can provide
            a single index or a range of indexes separated by a dash. For example,
            to extract volumes 0, 1, and 2, you can use '0-2'. You can also provide
            a list of indexes separated by a comma. For example, to extract volumes
            0, 1, 2, 8, 10, 11 and 12 , you can use '0-2,8,10-12'.
        grouped : bool, optional
            If True, extracted volumes are grouped into a single array. If False,
            save a list of separate volumes.
        out_dir : string, optional
            Output directory.
        out_vol : string, optional
            Name of the resulting volume.

        """
        io_it = self.get_io_iterator()
        vol_idx = handle_vol_idx(vol_idx)

        for fpath, ovol in io_it:
            logging.info("Extracting volume from {0}".format(fpath))
            data, affine, image = load_nifti(fpath, return_img=True)

            if grouped:
                split_vol = data[..., vol_idx]
                save_nifti(ovol, split_vol, affine, hdr=image.header)
                logging.info("Volume saved as {0}".format(ovol))
            else:
                for i in vol_idx:
                    fname = ovol.replace(".nii", f"_{i}.nii")
                    split_vol = data[..., i]
                    save_nifti(fname, split_vol, affine, hdr=image.header)
                    logging.info("Volume saved as {0}".format(fname))


class ConcatenateTractogramFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "concatracks"

    def run(
        self,
        tractogram_files,
        reference=None,
        delete_dpv=False,
        delete_dps=False,
        delete_groups=False,
        check_space_attributes=True,
        preallocation=False,
        out_dir="",
        out_extension="trx",
        out_tractogram="concatenated_tractogram",
    ):
        """Concatenate multiple tractograms into one.

        Parameters
        ----------
        tractogram_list : variable string
            The stateful tractogram filenames to concatenate
        reference : string, optional
            Reference anatomy for tck/vtk/fib/dpy file.
            support (.nii or .nii.gz).
        delete_dpv : bool, optional
            Delete dpv keys that do not exist in all the provided TrxFiles
        delete_dps : bool, optional
            Delete dps keys that do not exist in all the provided TrxFile
        delete_groups : bool, optional
            Delete all the groups that currently exist in the TrxFiles
        check_space_attributes : bool, optional
            Verify that dimensions and size of data are similar between all the
            TrxFiles
        preallocation : bool, optional
            Preallocated TrxFile has already been generated and is the first
            element in trx_list (Note: delete_groups must be set to True as
            well)
        out_dir : string, optional
            Output directory.
        out_extension : string, optional
            Extension of the resulting tractogram
        out_tractogram : string, optional
            Name of the resulting tractogram

        """
        io_it = self.get_io_iterator()

        trx_list = []
        has_group = False
        for fpath, _, _ in io_it:
            if fpath.lower().endswith(".trx") or fpath.lower().endswith(".trk"):
                reference = "same"

            if not reference:
                raise ValueError(
                    "No reference provided. It is needed for tck,"
                    "fib, dpy or vtk files"
                )

            tractogram_obj = load_tractogram(fpath, reference, bbox_valid_check=False)

            if not isinstance(tractogram_obj, tmm.TrxFile):
                tractogram_obj = tmm.TrxFile.from_sft(tractogram_obj)
            elif len(tractogram_obj.groups):
                has_group = True
            trx_list.append(tractogram_obj)

        trx = concatenate_tractogram(
            trx_list,
            delete_dpv=delete_dpv,
            delete_dps=delete_dps,
            delete_groups=delete_groups or not has_group,
            check_space_attributes=check_space_attributes,
            preallocation=preallocation,
        )

        valid_extensions = ["trk", "trx", "tck", "fib", "dpy", "vtk"]
        if out_extension.lower() not in valid_extensions:
            raise ValueError(
                f"Invalid extension. Valid extensions are: {valid_extensions}"
            )

        out_fpath = os.path.join(out_dir, f"{out_tractogram}.{out_extension}")
        save_tractogram(trx.to_sft(), out_fpath, bbox_valid_check=False)


class ConvertSHFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "convert_dipy_mrtrix"

    def run(
        self,
        input_files,
        out_dir="",
        out_file="sh_convert_dipy_mrtrix_out.nii.gz",
    ):
        """Converts SH basis representation between DIPY and MRtrix3 formats.
        Because this conversion is equal to its own inverse, it can be used to
        convert in either direction: DIPY to MRtrix3 or vice versa.

        Parameters
        ----------
        input_files : string
            Path to the input files. This path may contain wildcards to
            process multiple inputs at once.

        out_dir : string, optional
            Where the resulting file will be saved. (default '')

        out_file : string, optional
            Name of the result file to be saved.
            (default 'sh_convert_dipy_mrtrix_out.nii.gz')
        """

        io_it = self.get_io_iterator()

        for in_file, out_file in io_it:
            data, affine, image = load_nifti(in_file, return_img=True)
            data = convert_sh_descoteaux_tournier(data)
            save_nifti(out_file, data, affine, hdr=image.header)


class ConvertTensorsFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "convert_tensors"

    def run(
        self,
        tensor_files,
        from_format="mrtrix",
        to_format="dipy",
        out_dir=".",
        out_tensor="converted_tensor",
    ):
        """Converts tensor representation between different formats.

        Parameters
        ----------
        tensor_files : variable string
            Any number of tensor files
        from_format : string, optional
            Format of the input tensor files. Valid options are 'dipy',
            'mrtrix', 'ants', 'fsl'.
        to_format : string, optional
            Format of the output tensor files. Valid options are 'dipy',
            'mrtrix', 'ants', 'fsl'.
        out_dir : string, optional
            Output directory.
        out_tensor : string, optional
            Name of the resulting tensor file

        """
        io_it = self.get_io_iterator()
        for fpath, otensor in io_it:
            logging.info(f"Converting {fpath}")
            data, affine, image = load_nifti(fpath, return_img=True)
            data = convert_tensors(data, from_format, to_format)
            save_nifti(otensor, data, affine, hdr=image.header)


class ConvertTractogramFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "convert_tractogram"

    def run(
        self,
        input_files,
        reference=None,
        pos_dtype="float32",
        offsets_dtype="uint32",
        out_dir="",
        out_tractogram="converted_tractogram.trk",
    ):
        """Converts tractogram between different formats.

        Parameters
        ----------
        input_files : variable string
            Any number of tractogram files
        reference : string, optional
            Reference anatomy for tck/vtk/fib/dpy file.
            support (.nii or .nii.gz).
        pos_dtype : string, optional
            Data type of the tractogram points, used for vtk files.
        offsets_dtype : string, optional
            Data type of the tractogram offsets, used for vtk files.
        out_dir : string, optional
            Output directory.
        out_tractogram : string, optional
            Name of the resulting tractogram

        """
        io_it = self.get_io_iterator()

        for fpath, otracks in io_it:
            in_extension = fpath.lower().split(".")[-1]
            out_extension = otracks.lower().split(".")[-1]

            if in_extension == out_extension:
                warnings.warn(
                    "Input and output are the same file format. Skipping...",
                    stacklevel=2,
                )
                continue

            if not reference and in_extension in ["trx", "trk"]:
                reference = "same"

            if not reference and in_extension not in ["trx", "trk"]:
                raise ValueError(
                    "No reference provided. It is needed for tck,"
                    "fib, dpy or vtk files"
                )

            sft = load_tractogram(fpath, reference, bbox_valid_check=False)

            if out_extension != "trx":
                if out_extension == "vtk":
                    if sft.streamlines._data.dtype.name != pos_dtype:
                        sft.streamlines._data = sft.streamlines._data.astype(pos_dtype)
                    if offsets_dtype == "uint64" or offsets_dtype == "uint32":
                        offsets_dtype = offsets_dtype[1:]
                    if sft.streamlines._offsets.dtype.name != offsets_dtype:
                        sft.streamlines._offsets = sft.streamlines._offsets.astype(
                            offsets_dtype
                        )
                save_tractogram(sft, otracks, bbox_valid_check=False)
            else:
                trx = tmm.TrxFile.from_sft(sft)
                if trx.streamlines._data.dtype.name != pos_dtype:
                    trx.streamlines._data = trx.streamlines._data.astype(pos_dtype)
                if trx.streamlines._offsets.dtype.name != offsets_dtype:
                    trx.streamlines._offsets = trx.streamlines._offsets.astype(
                        offsets_dtype
                    )
                tmm.save(trx, otracks)
                trx.close()


class NiftisToPamFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "niftis_to_pam"

    def run(
        self,
        peaks_dir_files,
        peaks_values_files,
        peaks_indices_files,
        shm_files=None,
        gfa_files=None,
        sphere_files=None,
        default_sphere_name="repulsion724",
        out_dir="",
        out_pam="peaks.pam5",
    ):
        """Convert multiple nifti files to a single pam5 file.

        Parameters
        ----------
        peaks_dir_files : string
            Path to the input peaks directions volume. This path may contain
            wildcards to process multiple inputs at once.
        peaks_values_files : string
            Path to the input peaks values volume. This path may contain
            wildcards to process multiple inputs at once.
        peaks_indices_files : string
            Path to the input peaks indices volume. This path may contain
            wildcards to process multiple inputs at once.
        shm_files : string, optional
            Path to the input spherical harmonics volume. This path may
            contain wildcards to process multiple inputs at once.
        gfa_files : string, optional
            Path to the input generalized FA volume. This path may contain
            wildcards to process multiple inputs at once.
        sphere_files : string, optional
            Path to the input sphere vertices. This path may contain
            wildcards to process multiple inputs at once. If it is not define,
            default_sphere option will be used.
        default_sphere_name : string, optional
            Specify default sphere to use for spherical harmonics
            representation. This option can be superseded by
            sphere_files option. Possible options: ['symmetric362', 'symmetric642',
            'symmetric724', 'repulsion724', 'repulsion100', 'repulsion200'].
        out_dir : string, optional
            Output directory (default input file directory).
        out_pam : string, optional
            Name of the peaks volume to be saved.

        """
        io_it = self.get_io_iterator()

        msg = f"pam5 files saved in {out_dir or 'current directory'}"

        for fpeak_dirs, fpeak_values, fpeak_indices, opam in io_it:
            logging.info("Converting nifti files to pam5")
            peak_dirs, affine = load_nifti(fpeak_dirs)
            peak_values, _ = load_nifti(fpeak_values)
            peak_indices, _ = load_nifti(fpeak_indices)

            if sphere_files:
                xyz = np.loadtxt(sphere_files)
                sphere = Sphere(xyz=xyz)
            else:
                sphere = get_sphere(name=default_sphere_name)

            niftis_to_pam(
                affine=affine,
                peak_dirs=peak_dirs,
                sphere=sphere,
                peak_values=peak_values,
                peak_indices=peak_indices,
                pam_file=opam,
            )
            logging.info(msg.replace("pam5", opam))


class TensorToPamFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "tensor_to_niftis"

    def run(
        self,
        evals_files,
        evecs_files,
        sphere_files=None,
        default_sphere_name="repulsion724",
        out_dir="",
        out_pam="peaks.pam5",
    ):
        """Convert multiple tensor files(evals, evecs) to pam5 files.

        Parameters
        ----------
        evals_files : string
            Path to the input eigen values volumes. This path may contain
            wildcards to process multiple inputs at once.
        evecs_files : string
            Path to the input eigen vectors volumes. This path may contain
            wildcards to process multiple inputs at once.
        sphere_files : string, optional
            Path to the input sphere vertices. This path may contain
            wildcards to process multiple inputs at once. If it is not define,
            default_sphere option will be used.
        default_sphere_name : string, optional
            Specify default sphere to use for spherical harmonics
            representation. This option can be superseded by sphere_files
            option. Possible options: ['symmetric362', 'symmetric642',
            'symmetric724', 'repulsion724', 'repulsion100', 'repulsion200'].
        out_dir : string, optional
            Output directory (default input file directory).
        out_pam : string, optional
            Name of the peaks volume to be saved.

        """
        io_it = self.get_io_iterator()

        msg = f"pam5 files saved in {out_dir or 'current directory'}"

        for fevals, fevecs, opam in io_it:
            logging.info("Converting tensor files to pam5...")
            evals, affine = load_nifti(fevals)
            evecs, _ = load_nifti(fevecs)

            if sphere_files:
                xyz = np.loadtxt(sphere_files)
                sphere = Sphere(xyz=xyz)
            else:
                sphere = get_sphere(name=default_sphere_name)

            tensor_to_pam(evals, evecs, affine, sphere=sphere, pam_file=opam)
            logging.info(msg.replace("pam5", opam))


class PamToNiftisFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "pam_to_niftis"

    def run(
        self,
        pam_files,
        out_dir="",
        out_peaks_dir="peaks_dirs.nii.gz",
        out_peaks_values="peaks_values.nii.gz",
        out_peaks_indices="peaks_indices.nii.gz",
        out_shm="shm.nii.gz",
        out_gfa="gfa.nii.gz",
        out_sphere="sphere.txt",
        out_b="B.nii.gz",
        out_qa="qa.nii.gz",
    ):
        """Convert pam5 files to multiple nifti files.

        Parameters
        ----------
        pam_files : string
            Path to the input peaks volumes. This path may contain wildcards to
            process multiple inputs at once.
        out_dir : string, optional
            Output directory (default input file directory).
        out_peaks_dir : string, optional
            Name of the peaks directions volume to be saved.
        out_peaks_values : string, optional
            Name of the peaks values volume to be saved.
        out_peaks_indices : string, optional
            Name of the peaks indices volume to be saved.
        out_shm : string, optional
            Name of the spherical harmonics volume to be saved.
        out_gfa : string, optional
            Generalized FA volume name to be saved.
        out_sphere : string, optional
            Sphere vertices name to be saved.
        out_b : string, optional
            Name of the B Matrix to be saved.
        out_qa : string, optional
            Name of the Quantitative Anisotropy file to be saved.

        """
        io_it = self.get_io_iterator()

        msg = f"Nifti files saved in {out_dir or 'current directory'}"
        for (
            ipam,
            opeaks_dir,
            opeaks_values,
            opeaks_indices,
            oshm,
            ogfa,
            osphere,
            ob,
            oqa,
        ) in io_it:
            logging.info("Converting %s file to niftis...", ipam)
            pam = load_pam(ipam)
            pam_to_niftis(
                pam,
                fname_peaks_dir=opeaks_dir,
                fname_shm=oshm,
                fname_peaks_values=opeaks_values,
                fname_peaks_indices=opeaks_indices,
                fname_sphere=osphere,
                fname_gfa=ogfa,
                fname_b=ob,
                fname_qa=oqa,
            )
            logging.info(msg)


class MathFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "math_flow"

    def run(
        self, operation, input_files, dtype=None, out_dir="", out_file="math_out.nii.gz"
    ):
        """Perform mathematical operations on volume input files.

        This workflow allows the user to perform mathematical operations on
        multiple input files. e.g. to add two volumes together, subtract one:
        ``dipy_math "vol1 + vol2 - vol3" t1.nii.gz t1_a.nii.gz t1_b.nii.gz``
        The input files must be in Nifti format and have the same shape.

        Parameters
        ----------
        operation : string
            Mathematical operation to perform. supported operators are:
                - Bitwise operators (and, or, not, xor): ``&, |, ~, ^``
                - Comparison operators: ``<, <=, ==, !=, >=, >``
                - Unary arithmetic operators: ``-``
                - Binary arithmetic operators: ``+, -, *, /, **, <<, >>``
            Supported functions are:
                - ``where(bool, number1, number2) -> number``: number1 if the bool
                  condition is true, number2 otherwise.
                - ``{sin,cos,tan}(float|complex) -> float|complex``: trigonometric sine,
                  cosine or tangent.
                - ``{arcsin,arccos,arctan}(float|complex) -> float|complex``:
                  trigonometric inverse sine, cosine or tangent.
                - ``arctan2(float1, float2) -> float``: trigonometric inverse tangent of
                  float1/float2.
                - ``{sinh,cosh,tanh}(float|complex) -> float|complex``: hyperbolic
                  sine, cosine or tangent.
                - ``{arcsinh,arccosh,arctanh}(float|complex) -> float|complex``:
                  hyperbolic inverse sine, cosine or tangent.
                - ``{log,log10,log1p}(float|complex) -> float|complex``: natural,
                  base-10 and log(1+x) logarithms.
                - ``{exp,expm1}(float|complex) -> float|complex``: exponential and
                  exponential minus one.
                - ``sqrt(float|complex) -> float|complex``: square root.
                - ``abs(float|complex) -> float|complex``: absolute value.
                - ``conj(complex) -> complex``: conjugate value.
                - ``{real,imag}(complex) -> float``: real or imaginary part of complex.
                - ``complex(float, float) -> complex``: complex from real and imaginary
                  parts.
                - ``contains(np.str, np.str) -> bool``: returns True for every string
                  in op1 that contains op2.
        input_files : variable string
            Any number of Nifti1 files
        dtype : string, optional
            Data type of the resulting file.
        out_dir : string, optional
            Output directory
        out_file : string, optional
            Name of the resulting file to be saved.
        """
        vol_dict = {}
        ref_affine = None
        ref_shape = None
        info_msg = ""
        have_errors = False
        for i, fname in enumerate(input_files, start=1):
            if not os.path.isfile(fname):
                logging.error(f"Input file {fname} does not exist.")
                raise SystemExit()

            if not (fname.endswith(".nii.gz") or fname.endswith(".nii")):
                msg = (
                    f"Wrong volume type: {fname}. Only Nifti files are supported"
                    " (*.nii or *.nii.gz)."
                )
                logging.error(msg)
                raise SystemExit()

            data, affine = load_nifti(fname)
            vol_dict[f"vol{i}"] = data
            info_msg += f"{fname}:\n- vol index: {i}\n- shape: {data.shape}"
            info_msg += f"\n- affine:\n{affine}\n"
            if ref_affine is None:
                ref_affine = affine
                ref_shape = data.shape
                continue

            have_errors = (
                have_errors
                or not np.all(np.isclose(ref_affine, affine, rtol=1e-05, atol=1e-08))
                or not np.array_equal(ref_shape, data.shape)
            )

        if have_errors:
            logging.warning(info_msg)
            msg = "All input files must have the same shape and affine matrix."
            logging.error(msg)
            raise SystemExit()

        try:
            res = ne.evaluate(operation, local_dict=vol_dict)
        except KeyError as e:
            msg = (
                f"Impossible key {e} in the operation. You have {len(input_files)}"
                f" volumes available with the following keys: {list(vol_dict.keys())}"
            )
            logging.error(msg)
            raise SystemExit() from e

        if dtype:
            try:
                res = res.astype(dtype)
            except TypeError as e:
                msg = (
                    f"Impossible to cast to {dtype}. Check possible numpy type here:"
                    "https://numpy.org/doc/stable/reference/arrays.interface.html"
                )
                logging.error(msg)
                raise SystemExit() from e

        out_fname = os.path.join(out_dir, out_file)
        logging.info(f"Saving result to {out_fname}")
        save_nifti(out_fname, res, affine)



import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

matplotlib, has_mpl, setup_module = optional_package("matplotlib")
plt, _, _ = optional_package("matplotlib.pyplot")


def _tile_plot(imgs, titles, **kwargs):
    """
    Helper function
    """
    # Create a new figure and plot the three images
    fig, ax = plt.subplots(1, len(imgs))
    for ii, a in enumerate(ax):
        a.set_axis_off()
        a.imshow(imgs[ii], **kwargs)
        a.set_title(titles[ii])

    return fig


def simple_plot(file_name, title, x, y, xlabel, ylabel):
    """Saves the simple plot with given x and y values

    Parameters
    ----------
    file_name : string
        file name for saving the plot
    title : string
        title of the plot
    x : integer list
        x-axis values to be plotted
    y : integer list
        y-axis values to be plotted
    xlabel : string
        label for x-axis
    ylable : string
        label for y-axis

    """

    plt.plot(x, y)
    axes = plt.gca()
    axes.set_ylim([0, 4])
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)

    plt.savefig(file_name)
    plt.clf()


@warning_for_keywords()
def overlay_images(
    img0, img1, *, title0="", title_mid="", title1="", fname=None, **fig_kwargs
):
    r"""Plot two images one on top of the other using red and green channels.

    Creates a figure containing three images: the first image to the left
    plotted on the red channel of a color image, the second to the right
    plotted on the green channel of a color image and the two given images on
    top of each other using the red channel for the first image and the green
    channel for the second one. It is assumed that both images have the same
    shape. The intended use of this function is to visually assess the quality
    of a registration result.

    Parameters
    ----------
    img0 : array, shape(R, C)
        the image to be plotted on the red channel, to the left of the figure
    img1 : array, shape(R, C)
        the image to be plotted on the green channel, to the right of the
        figure
    title0 : string, optional
        the title to be written on top of the image to the left. By default, no
        title is displayed.
    title_mid : string, optional
        the title to be written on top of the middle image. By default, no
        title is displayed.
    title1 : string, optional
        the title to be written on top of the image to the right. By default,
        no title is displayed.
    fname : string, optional
        the file name to write the resulting figure. If None (default), the
        image is not saved.
    fig_kwargs : dict
        Extra parameters for saving figure, e.g. `dpi=300`.
    """
    # Normalize the input images to [0,255]
    img0 = 255 * ((img0 - img0.min()) / (img0.max() - img0.min()))
    img1 = 255 * ((img1 - img1.min()) / (img1.max() - img1.min()))

    # Create the color images
    img0_red = np.zeros(shape=img0.shape + (3,), dtype=np.uint8)
    img1_green = np.zeros(shape=img0.shape + (3,), dtype=np.uint8)
    overlay = np.zeros(shape=img0.shape + (3,), dtype=np.uint8)

    # Copy the normalized intensities into the appropriate channels of the
    # color images
    img0_red[..., 0] = img0
    img1_green[..., 1] = img1
    overlay[..., 0] = img0
    overlay[..., 1] = img1

    fig = _tile_plot([img0_red, overlay, img1_green], [title0, title_mid, title1])

    # If a file name was given, save the figure
    if fname is not None:
        fig.savefig(fname, bbox_inches="tight", **fig_kwargs)

    return fig


def draw_lattice_2d(nrows, ncols, delta):
    r"""Create a regular lattice of nrows x ncols squares.

    Creates an image (2D array) of a regular lattice of nrows x ncols squares.
    The size of each square is delta x delta pixels (not counting the
    separation lines). The lines are one pixel width.

    Parameters
    ----------
    nrows : int
        the number of squares to be drawn vertically
    ncols : int
        the number of squares to be drawn horizontally
    delta : int
        the size of each square of the grid. Each square is delta x delta
        pixels

    Returns
    -------
    lattice : array, shape (R, C)
        the image (2D array) of the segular lattice. The shape (R, C) of the
        array is given by
        R = 1 + (delta + 1) * nrows
        C = 1 + (delta + 1) * ncols
    """
    lattice = np.ndarray(
        (1 + (delta + 1) * nrows, 1 + (delta + 1) * ncols), dtype=np.float64
    )

    # Fill the lattice with "white"
    lattice[...] = 127

    # Draw the horizontal lines in "black"
    for i in range(nrows + 1):
        lattice[i * (delta + 1), :] = 0

    # Draw the vertical lines in "black"
    for j in range(ncols + 1):
        lattice[:, j * (delta + 1)] = 0

    return lattice


@warning_for_keywords()
def plot_2d_diffeomorphic_map(
    mapping,
    *,
    delta=10,
    fname=None,
    direct_grid_shape=None,
    direct_grid2world=-1,
    inverse_grid_shape=None,
    inverse_grid2world=-1,
    show_figure=True,
    **fig_kwargs,
):
    r"""Draw the effect of warping a regular lattice by a diffeomorphic map.

    Draws a diffeomorphic map by showing the effect of the deformation on a
    regular grid. The resulting figure contains two images: the direct
    transformation is plotted to the left, and the inverse transformation is
    plotted to the right.

    Parameters
    ----------
    mapping : DiffeomorphicMap object
        the diffeomorphic map to be drawn
    delta : int, optional
        the size (in pixels) of the squares of the regular lattice to be used
        to plot the warping effects. Each square will be delta x delta pixels.
        By default, the size will be 10 pixels.
    fname : string, optional
        the name of the file the figure will be written to. If None (default),
        the figure will not be saved to disk.
    direct_grid_shape : tuple, shape (2,), optional
        the shape of the grid image after being deformed by the direct
        transformation. By default, the shape of the deformed grid is the
        same as the grid of the displacement field, which is by default
        equal to the shape of the fixed image. In other words, the resulting
        deformed grid (deformed by the direct transformation) will normally
        have the same shape as the fixed image.
    direct_grid2world : array, shape (3, 3), optional
        the affine transformation mapping the direct grid's coordinates to
        physical space. By default, this transformation will correspond to
        the image-to-world transformation corresponding to the default
        direct_grid_shape (in general, if users specify a direct_grid_shape,
        they should also specify direct_grid2world).
    inverse_grid_shape : tuple, shape (2,), optional
        the shape of the grid image after being deformed by the inverse
        transformation. By default, the shape of the deformed grid under the
        inverse transform is the same as the image used as "moving" when
        the diffeomorphic map was generated by a registration algorithm
        (so it corresponds to the effect of warping the static image towards
        the moving).
    inverse_grid2world : array, shape (3, 3), optional
        the affine transformation mapping inverse grid's coordinates to
        physical space. By default, this transformation will correspond to
        the image-to-world transformation corresponding to the default
        inverse_grid_shape (in general, if users specify an inverse_grid_shape,
        they should also specify inverse_grid2world).
    show_figure : bool, optional
        if True (default), the deformed grids will be plotted using matplotlib,
        else the grids are just returned
    fig_kwargs : dict
        Extra parameters for saving figure, e.g. `dpi=300`.

    Returns
    -------
    warped_forward : array
        Image with the grid showing the effect of transforming the moving image to
        the static image.  The shape will be `direct_grid_shape` if specified,
        otherwise the shape of the static image.
    warped_backward : array
        Image with the grid showing the effect of transforming the static image to
        the moving image.  Shape will be `inverse_grid_shape` if specified,
        otherwise the shape of the moving image.

    Notes
    -----
    The default value for the affine transformation is "-1" to handle the case
    in which the user provides "None" as input meaning "identity". If we used
    None as default, we wouldn't know if the user specifically wants to use
    the identity (specifically passing None) or if it was left unspecified,
    meaning to use the appropriate default matrix.
    """
    if mapping.is_inverse:
        # By default, direct_grid_shape is the codomain grid
        if direct_grid_shape is None:
            direct_grid_shape = mapping.codomain_shape
        if direct_grid2world == -1:
            direct_grid2world = mapping.codomain_grid2world

        # By default, the inverse grid is the domain grid
        if inverse_grid_shape is None:
            inverse_grid_shape = mapping.domain_shape
        if inverse_grid2world == -1:
            inverse_grid2world = mapping.domain_grid2world
    else:
        # Now by default, direct_grid_shape is the mapping's input grid
        if direct_grid_shape is None:
            direct_grid_shape = mapping.domain_shape
        if direct_grid2world == -1:
            direct_grid2world = mapping.domain_grid2world

        # By default, the output grid is the mapping's domain grid
        if inverse_grid_shape is None:
            inverse_grid_shape = mapping.codomain_shape
        if inverse_grid2world == -1:
            inverse_grid2world = mapping.codomain_grid2world

    # The world-to-image (image = drawn lattice on the output grid)
    # transformation is the inverse of the output affine
    world_to_image = None
    if inverse_grid2world is not None:
        world_to_image = np.linalg.inv(inverse_grid2world)

    # Draw the squares on the output grid
    lattice_out = draw_lattice_2d(
        (inverse_grid_shape[0] + delta) // (delta + 1),
        (inverse_grid_shape[1] + delta) // (delta + 1),
        delta,
    )
    lattice_out = lattice_out[0 : inverse_grid_shape[0], 0 : inverse_grid_shape[1]]

    # Warp in the forward direction (sampling it on the input grid)
    warped_forward = mapping.transform(
        lattice_out,
        interpolation="linear",
        image_world2grid=world_to_image,
        out_shape=direct_grid_shape,
        out_grid2world=direct_grid2world,
    )

    # Now, the world-to-image (image = drawn lattice on the input grid)
    # transformation is the inverse of the input affine
    world_to_image = None
    if direct_grid2world is not None:
        world_to_image = np.linalg.inv(direct_grid2world)

    # Draw the squares on the input grid
    lattice_in = draw_lattice_2d(
        (direct_grid_shape[0] + delta) // (delta + 1),
        (direct_grid_shape[1] + delta) // (delta + 1),
        delta,
    )
    lattice_in = lattice_in[0 : direct_grid_shape[0], 0 : direct_grid_shape[1]]

    # Warp in the backward direction (sampling it on the output grid)
    warped_backward = mapping.transform_inverse(
        lattice_in,
        interpolation="linear",
        image_world2grid=world_to_image,
        out_shape=inverse_grid_shape,
        out_grid2world=inverse_grid2world,
    )

    # Now plot the grids
    if show_figure:
        plt.figure()

        plt.subplot(1, 2, 1).set_axis_off()
        plt.imshow(warped_forward, cmap=plt.cm.gray)
        plt.title("Direct transform")

        plt.subplot(1, 2, 2).set_axis_off()
        plt.imshow(warped_backward, cmap=plt.cm.gray)
        plt.title("Inverse transform")

        # Finally, save the figure to disk
        if fname is not None:
            plt.savefig(fname, bbox_inches="tight", **fig_kwargs)

    # Return the deformed grids
    return warped_forward, warped_backward


@warning_for_keywords()
def plot_slices(V, *, slice_indices=None, fname=None, **fig_kwargs):
    r"""Plot 3 slices from the given volume: 1 sagittal, 1 coronal and 1 axial

    Creates a figure showing the axial, coronal and sagittal slices at the
    requested positions of the given volume. The requested slices are specified
    by slice_indices.

    Parameters
    ----------
    V : array, shape (S, R, C)
        the 3D volume to extract the slices from
    slice_indices : array, shape (3,), optional
        the indices of the sagittal (slice_indices[0]), coronal
        (slice_indices[1])
        and axial (slice_indices[2]) slices to be displayed. If None, the
        middle slices along each direction are displayed.
    fname : string, optional
        the name of the file to save the figure to. If None (default), the
        figure is not saved to disk.
    fig_kwargs : dict
        Extra parameters for saving figure, e.g. `dpi=300`.
    """
    if slice_indices is None:
        slice_indices = np.array(V.shape) // 2

    # Normalize the intensities to [0, 255]
    V = np.asarray(V, dtype=np.float64)
    V = 255 * (V - V.min()) / (V.max() - V.min())

    # Extract the middle slices
    axial = np.asarray(V[:, :, slice_indices[2]]).astype(np.uint8).T
    coronal = np.asarray(V[:, slice_indices[1], :]).astype(np.uint8).T
    sagittal = np.asarray(V[slice_indices[0], :, :]).astype(np.uint8).T

    fig = _tile_plot(
        [axial, coronal, sagittal],
        ["Axial", "Coronal", "Sagittal"],
        cmap=plt.cm.gray,
        origin="lower",
    )

    # Save the figure if requested
    if fname is not None:
        fig.savefig(fname, bbox_inches="tight", **fig_kwargs)

    return fig


@warning_for_keywords()
def overlay_slices(
    L,
    R,
    *,
    slice_index=None,
    slice_type=1,
    ltitle="Left",
    rtitle="Right",
    fname=None,
    **fig_kwargs,
):
    r"""Plot three overlaid slices from the given volumes.

    Creates a figure containing three images: the gray scale k-th slice of
    the first volume (L) to the left, where k=slice_index, the k-th slice of
    the second volume (R) to the right and the k-th slices of the two given
    images on top of each other using the red channel for the first volume and
    the green channel for the second one. It is assumed that both volumes have
    the same shape. The intended use of this function is to visually assess the
    quality of a registration result.

    Parameters
    ----------
    L : array, shape (S, R, C)
        the first volume to extract the slice from plotted to the left
    R : array, shape (S, R, C)
        the second volume to extract the slice from, plotted to the right
    slice_index : int, optional
        the index of the slices (along the axis given by slice_type) to be
        overlaid. If None, the slice along the specified axis is used
    slice_type : int, optional
        the type of slice to be extracted:
        0=sagittal, 1=coronal (default), 2=axial.
    ltitle : string, optional
        the string to be written as the title of the left image. By default,
        no title is displayed.
    rtitle : string, optional
        the string to be written as the title of the right image. By default,
        no title is displayed.
    fname : string, optional
        the name of the file to write the image to. If None (default), the
        figure is not saved to disk.
    fig_kwargs: extra parameters for saving figure, e.g. `dpi=300`.
    """

    # Normalize the intensities to [0,255]
    sh = L.shape
    L = np.asarray(L, dtype=np.float64)
    R = np.asarray(R, dtype=np.float64)
    L = 255 * (L - L.min()) / (L.max() - L.min())
    R = 255 * (R - R.min()) / (R.max() - R.min())

    # Create the color image to draw the overlapped slices into, and extract
    # the slices (note the transpositions)
    if slice_type == 0:
        if slice_index is None:
            slice_index = sh[0] // 2
        colorImage = np.zeros(shape=(sh[2], sh[1], 3), dtype=np.uint8)
        ll = np.asarray(L[slice_index, :, :]).astype(np.uint8).T
        rr = np.asarray(R[slice_index, :, :]).astype(np.uint8).T
    elif slice_type == 1:
        if slice_index is None:
            slice_index = sh[1] // 2
        colorImage = np.zeros(shape=(sh[2], sh[0], 3), dtype=np.uint8)
        ll = np.asarray(L[:, slice_index, :]).astype(np.uint8).T
        rr = np.asarray(R[:, slice_index, :]).astype(np.uint8).T
    elif slice_type == 2:
        if slice_index is None:
            slice_index = sh[2] // 2
        colorImage = np.zeros(shape=(sh[1], sh[0], 3), dtype=np.uint8)
        ll = np.asarray(L[:, :, slice_index]).astype(np.uint8).T
        rr = np.asarray(R[:, :, slice_index]).astype(np.uint8).T
    else:
        print("Slice type must be 0, 1 or 2.")
        return

    # Draw the intensity images to the appropriate channels of the color image
    # The "(ll > ll[0, 0])" condition is just an attempt to eliminate the
    # background when its intensity is not exactly zero (the [0,0] corner is
    # usually background)
    colorImage[..., 0] = ll * (ll > ll[0, 0])
    colorImage[..., 1] = rr * (rr > rr[0, 0])

    fig = _tile_plot(
        [ll, colorImage, rr],
        [ltitle, "Overlay", rtitle],
        cmap=plt.cm.gray,
        origin="lower",
    )

    # Save the figure to disk, if requested
    if fname is not None:
        fig.savefig(fname, bbox_inches="tight", **fig_kwargs)

    return fig
